(* Author: Dmitriy Traytel *)

theory M2L_Examples
imports M2L_Equivalence_Checking "~~/src/HOL/Library/Char_ord"
begin

lemma "check_eqv' [True, False] 0 (FExists (FEXISTS (FIn 1 0))) (FEXISTS (FExists (FIn 0 1)))"
  by eval

abbreviation FALSE where "FALSE \<equiv> FExists (FLess 0 0)"
abbreviation TRUE where "TRUE \<equiv> FNot FALSE"
abbreviation All where "All \<phi> \<equiv> FNot (FExists (FNot \<phi>))"
abbreviation AL where "AL \<phi> \<equiv> FNot (FEXISTS (FNot \<phi>))"
abbreviation Imp where "Imp \<phi> \<psi> \<equiv> FOr (FNot \<phi>) \<psi>"
abbreviation Iff where "Iff \<phi> \<psi> \<equiv> FAnd (Imp \<phi> \<psi>) (Imp \<psi> \<phi>)"
abbreviation Ball where "Ball X \<phi> \<equiv> All (Imp (FIn 0 (X+1)) \<phi>)"
abbreviation Bex where "Bex X \<phi> \<equiv> FExists (FAnd (FIn 0 (X+1)) \<phi>)"

abbreviation Eq where "Eq x y \<equiv> FAnd (FNot (FLess x y)) (FNot (FLess y x))"
abbreviation SUBSET where "SUBSET X Y \<equiv> Ball X (FIn 0 (Y+1))"
abbreviation EQ where "EQ X Y \<equiv> FAnd (SUBSET X Y) (SUBSET Y X)"
abbreviation First where "First x \<equiv> FNot (FExists (FLess 0 (x+1)))"
abbreviation Last where "Last x \<equiv> FNot (FExists (FLess (x+1) 0))"
abbreviation Suc where "Suc sucx x \<equiv> FAnd (FLess x sucx) (FNot (FExists (FAnd (FLess (x+1) 0) (FLess 0 (sucx+1)))))"

(*definition "Thm \<Sigma> n \<phi> = check_eqv \<Sigma> n \<phi> TRUE"*)
definition "Thm \<Sigma> n \<phi> = check_eqv' \<Sigma> n \<phi> TRUE"

(*export_code Thm in SML module_name Thm (*"file "Thm.ML"*)*)

lemma "Thm [True] 0 (FEXISTS (All (FIn 0 1)))"
  by eval

definition "mod_two a b c d = Iff a (Iff b (Iff c d))"
definition "at_least_two a b c d = Iff d (FOr (FAnd a b) (FOr (FAnd b c) (FAnd a c)))"

definition "ADD S A B = FEXISTS (FAnd (All (Imp (First 0) (FNot (FIn 0 1))))
  (All (FAnd (mod_two (FIn 0 (A+2)) (FIn 0 (B+2)) (FIn 0 1) (FIn 0 (S+2)))
       (Imp (FExists (FAnd (Last 0) (FLess 1 0)))
            (at_least_two (FIn 0 (A+2)) (FIn 0 (B+2)) (FIn 0 1) (All (Imp (Suc 0 1) (FIn 0 2))))))))"

definition Comm_Lemma :: "bool formula" where
  "Comm_Lemma = AL (AL (FEXISTS (FAnd (ADD 0 1 2) (FEXISTS (FAnd (ADD 0 3 2) (EQ 0 1))))))"
(*
export_code rexp.rexp_size in SML module_name Thm

export_code Comm_Lemma Thm in SML module_name Thm (*file "Thm.ML"*) 
*)
(*
lemma "Thm [True] 0 (AL (AL (EXISTS (ADD 0 1 2))))"
  by eval

lemma "Thm [True] 0 Comm_Lemma"
  by eval
*)
(*
value[code] "Thm [True, False] 0 (AL (AL (EXISTS (EXISTS (And (Exists (And (First 0) (Not (In 0 1))))
  (All ((mod_two (In 0 4) (In 0 3) (In 0 1) (In 0 2)))))))))"
*)
abbreviation Globally ("\<box>_" [40] 40) where "Globally P == %n. All (Imp (FNot (FLess (n+1) 0)) (P 0))"
abbreviation Future ("\<diamondsuit>_" [40] 40) where "Future P == %n. FExists (FAnd (FNot (FLess (n+1) 0)) (P 0))"
abbreviation IMP (infixr "\<rightarrow>" 50) where "IMP P1 P2 == %n. Imp (P1 n) (P2 n)"

declare [[show_abbrevs=false]]
value "norm(All (((\<box>(%n. FQ True n)) \<rightarrow> (\<box>\<diamondsuit>(%n. FQ True n))) 0))"

primrec print_bool where
  "print_bool True = ''a''"
| "print_bool False = ''b''"

primrec print_rexp where
  "print_rexp Zero = ''0''"
| "print_rexp One = ''1''"
| "print_rexp (Atom x) = print_bool (fst x) @ '':'' @ List.maps print_bool (snd x)"
| "print_rexp (Plus r s) = ''('' @ print_rexp r @ ''+'' @ print_rexp s @ '')''"
| "print_rexp (Times r s) = print_rexp r @ print_rexp s"
| "print_rexp (Inter r s) = ''('' @ print_rexp r @ ''n'' @ print_rexp s @ '')''"
| "print_rexp (Star r) = ''('' @ print_rexp r @ '')*''"
| "print_rexp (rexp.Not r) = ''~('' @ print_rexp r @ '')''"
| "print_rexp (Pr r) = ''pi('' @ print_rexp r @ '')''"

value "rexp.rexp_size (\<lambda>_. 1) (PLUS [Atom ''a'', Zero, One])"
value "rexp.rexp_size (\<lambda>(_, bs). length bs) (Smart_Constructors_Normalization.norm
  (rexp_of [True, False] 1 (norm (All (((\<box>(%n. FQ (Some True) n)) \<rightarrow> (\<box>\<diamondsuit>(%n. FQ (Some True) n))) 0)))))"

value "rexp.rexp_size (\<lambda>(_, bs). length bs) (Smart_Constructors_Normalization.norm
  (rexp_of' [True, False] 1 (norm (All (((\<box>(%n. FQ (Some True) n)) \<rightarrow> (\<box>\<diamondsuit>(%n. FQ (Some True) n))) 0)))))"
(*
value "rexp.rexp_size (\<lambda>(_, bs). length bs) (Smart_Constructors_Normalization.norm
  (rexp_of'' [True, False] 1 (norm (All (((\<box>(%n. FQ True n)) \<rightarrow> (\<box>\<diamondsuit>(%n. FQ True n))) 0)))))"
*)
lemma "Thm [True, False] 0 (Not (Exists (Exists (And (Less 0 1) (Less 1 0)))))"
  by eval

lemma Drinker: "Thm [True, False] 1 (Exists (Imp (In 0 1) (All (In 0 2))))"
  by eval

lemma "Thm [True, False] 1 (Imp (All (In 0 1)) (Exists (In 0 1)))"
  by eval

lemma "Thm [True] 1 (All (((\<box>(%n. Q True n)) \<rightarrow> (\<diamondsuit>(%n. Q True n))) 0))"
  by eval

lemma "Thm [True] 0 (All (((\<box>(%n. Q True n)) \<rightarrow> (\<box>\<diamondsuit>(%n. Q True n))) 0))"
  by eval

lemma "Thm [True, False] 1 (All (((\<box>(%n. In n 2)) \<rightarrow> (\<diamondsuit>(%n. In n 2))) 0))"
  by eval

lemma "Thm [True, False] 0 (All (((\<box>(%n. Q True n)) \<rightarrow> (\<diamondsuit>(%n. Q True n))) 0))"
  by eval

lemma "Thm [True, False] 0 (All (((\<box>(%n. Q True n)) \<rightarrow> (\<box>\<diamondsuit>(%n. Q True n))) 0))"
  by eval

lemma "Thm [True, False] 1 (All (((\<box>(%n. In n 2)) \<rightarrow> (\<box>\<diamondsuit>(%n. In n 3))) 0))"
  by eval
(*
lemma "Thm [(''P1'', True), (''P1'', False), (''P2'', True), (''P2'', False)] 0
  (All (((\<box>(%n. Imp (Q (''P1'', True) n) (Q (''P2'', True) n))) \<rightarrow>
        (\<box>(%n. Q (''P1'', True) n)) \<rightarrow> (\<box>(%n. Q (''P2'', True) n))) 0))"
  by eval

lemma "Thm [(''P1'', True), (''P1'', False), (''P2'', True), (''P2'', False),
            (''P3'', True), (''P3'', False)] 0
  (All (((\<box>(%n. Imp (Q (''P1'', True) n) (Imp (Q (''P2'', True) n) (Q (''P3'', True) n)))) \<rightarrow>
        (\<box>(%n. Q (''P1'', True) n)) \<rightarrow> (\<box>(%n. Q (''P2'', True) n)) \<rightarrow> (\<box>(%n. Q (''P3'', True) n))) 0))"
  by eval

lemma "Thm [(''P1'', True), (''P1'', False), (''P2'', True), (''P2'', False),
            (''P3'', True), (''P3'', False), (''P4'', True), (''P4'', False)] 0
  (All (((\<box>(%n. Imp (Q (''P1'', True) n) (Imp (Q (''P2'', True) n) (Imp (Q (''P3'', True) n) (Q (''P4'', True) n))))) \<rightarrow>
        (\<box>(%n. Q (''P1'', True) n)) \<rightarrow> (\<box>(%n. Q (''P2'', True) n)) \<rightarrow> (\<box>(%n. Q (''P3'', True) n)) \<rightarrow> (\<box>(%n. Q (''P4'', True) n))) 0))"
  by eval
*)

abbreviation "OR xs n \<equiv> rexp_of_list Or FALSE (map (\<lambda>x. Q x n) xs)"

abbreviation \<Phi>0 :: "bool formula" where "\<Phi>0 \<equiv>
  (All (((\<box>(OR [(True)])) \<rightarrow> (\<box>(OR [(True)]))) 0))"

abbreviation \<Psi>0 :: "bool formula" where "\<Psi>0 \<equiv>
  (All (((\<box>(\<lambda>n. In n 2)) \<rightarrow> (\<box>(\<lambda>n. In n 2))) 0))"

abbreviation \<Phi>1 :: "(bool \<times> bool) formula" where "\<Phi>1 \<equiv>
  (All (((\<box>(OR [(True, True), (True, False)] \<rightarrow>
            OR [(True, True), (False, True)])) \<rightarrow>
         (\<box>(OR [(True, True), (True, False)])) \<rightarrow>
         (\<box>(OR [(True, True), (False, True)]))) 0))"

abbreviation \<Psi>1 :: "bool formula" where "\<Psi>1 \<equiv>
  (All (((\<box>((\<lambda>n. In n 2) \<rightarrow> (\<lambda>n. In n 3))) \<rightarrow> (\<box>(\<lambda>n. In n 2)) \<rightarrow> (\<box>(\<lambda>n. In n 3))) 0))"

abbreviation \<Phi>2 :: "(bool \<times> bool \<times> bool) formula" where "\<Phi>2 \<equiv> 
  (All (((\<box>(OR [(True, True, True), (True, True, False), (True, False, True), (True, False, False)] \<rightarrow>
            OR [(True, True, True), (True, True, False), (False, True, True), (False, True, False)] \<rightarrow>
            OR [(True, True, True), (True, False, True), (False, True, True), (False, False, True)])) \<rightarrow>
        (\<box>(OR [(True, True, True), (True, True, False), (True, False, True), (True, False, False)])) \<rightarrow>
        (\<box>(OR [(True, True, True), (True, True, False), (False, True, True), (False, True, False)])) \<rightarrow>
        (\<box>(OR [(True, True, True), (True, True, False), (False, True, True), (False, True, False)]))) 0))"

abbreviation \<Psi>2 :: "bool formula" where "\<Psi>2 \<equiv>
  (All (((\<box>((\<lambda>n. In n 2) \<rightarrow> (\<lambda>n. In n 3) \<rightarrow> (\<lambda>n. In n 4))) \<rightarrow>
    (\<box>(\<lambda>n. In n 2)) \<rightarrow> (\<box>(\<lambda>n. In n 3)) \<rightarrow> (\<box>(\<lambda>n. In n 4))) 0))"

value "rexp.rexp_size (\<lambda>(_, bs). length bs) ((rexp_of' [True, False] 0 (norm \<Phi>0)))"
value "rexp.rexp_size (\<lambda>(_, bs). length bs) ((rexp_of' [True] 1 (norm \<Psi>0)))"
value "rexp.rexp_size (\<lambda>(_, bs). 1) ((rexp_of' [True, False] 0 (norm \<Phi>0)))"
value "rexp.rexp_size (\<lambda>(_, bs). 1) ((rexp_of' [True] 1 (norm \<Psi>0)))"

value "rexp.rexp_size (\<lambda>(_, bs). length bs) ((rexp_of' [(True, True), (True, False), (False, True), (False, False)] 0 (norm \<Phi>1)))"
value "rexp.rexp_size (\<lambda>(_, bs). length bs) ((rexp_of' [True] 2 (norm \<Psi>1)))"
value "rexp.rexp_size (\<lambda>(_, bs). 1) ((rexp_of' [(True, True), (True, False), (False, True), (False, False)] 0 (norm \<Phi>1)))"
value "rexp.rexp_size (\<lambda>(_, bs). 1) ((rexp_of' [True] 2 (norm \<Psi>2)))"

value "rexp.rexp_size (\<lambda>(_, bs). length bs) ((rexp_of' [(True, True, True), (True, True, False), (True, False, True), (True, False, False),
            (False, True, True), (False, True, False), (False, False, True), (False, False, False)] 0 (norm \<Phi>2)))"
value "rexp.rexp_size (\<lambda>(_, bs). length bs) ((rexp_of' [True] 3 (norm \<Psi>2)))"
value "rexp.rexp_size (\<lambda>(_, bs). 1) ((rexp_of' [(True, True, True), (True, True, False), (True, False, True), (True, False, False),
            (False, True, True), (False, True, False), (False, False, True), (False, False, False)] 0 (norm \<Phi>2)))"
value "rexp.rexp_size (\<lambda>(_, bs). 1) ((rexp_of' [True] 3 (norm \<Psi>2)))"




lemma "Thm [True] 1 \<Psi>0"
  by eval
lemma "Thm [True] 2 \<Psi>1"
  by eval
lemma "Thm [True] 3 \<Psi>2"
  by eval

lemma "Thm [True, False] 0 \<Phi>0"
  by eval
lemma "Thm [(True, True), (True, False), (False, True), (False, False)] 0 \<Phi>1"
  by eval
lemma "Thm [(True, True, True), (True, True, False), (True, False, True), (True, False, False),
            (False, True, True), (False, True, False), (False, False, True), (False, False, False)] 0 \<Phi>2"
  by eval

end
