(* Author: Dmitriy Traytel *)

header {* Deciding Equivalence of MSO Formulas *}

(*<*)
theory MSO_Equivalence_Checking
imports MSO_Normalization Smart_Constructors_Normalization
begin
(*>*)

type_synonym 'a T = "'a \<times> bool list"
abbreviation "\<epsilon> \<equiv> \<lambda>\<Sigma> (a::'a, bs). if a \<in> set \<Sigma> then [(a, True # bs), (a, False # bs)] else []"
abbreviation "\<LL> \<equiv> \<lambda>\<Sigma>. project.lang (set o (\<sigma> \<Sigma>)) \<pi>"

definition wf_rexp where [code del]:
  "wf_rexp \<Sigma> = alphabet.wf (set o \<sigma> \<Sigma>)"

interpretation project "set o \<sigma> \<Sigma>" \<pi>
  where "alphabet.wf (set o \<sigma> \<Sigma>) = wf_rexp \<Sigma>"
by (unfold_locales) (auto simp: \<sigma>_def \<pi>_def wf_rexp_def)

definition norm_ederiv where [code del]:
  "norm_ederiv \<equiv> \<lambda>\<Sigma>. embed.ederiv (\<epsilon> \<Sigma>)"

interpretation embed "set \<circ> (\<sigma> (\<Sigma> :: 'a :: linorder list))" \<pi> "\<epsilon> \<Sigma>"
  where "embed.ederiv (\<epsilon> \<Sigma>) = norm_ederiv \<Sigma>"
by (unfold_locales) (auto simp: norm_ederiv_def \<sigma>_def \<pi>_def)

definition norm_step' where [code del]:
  "norm_step' \<equiv> \<lambda>\<Sigma>. (*efficient_*)equivalence_checker.step' (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_closure' where [code del]:
  "norm_closure' \<equiv> \<lambda>\<Sigma>. (*efficient_*)equivalence_checker.closure' (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_check_eqv' where [code del]:
  "norm_check_eqv' \<equiv> \<lambda>\<Sigma>. (*efficient_*)equivalence_checker.check_eqv' (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_step where [code del]:
  "norm_step \<equiv> \<lambda>\<Sigma>. (*efficient_*)equivalence_checker.step (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_closure where [code del]:
  "norm_closure \<equiv> \<lambda>\<Sigma>. (*efficient_*)equivalence_checker.closure (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_check_eqv where [code del]:
  "norm_check_eqv \<equiv> \<lambda>\<Sigma>. (*efficient_*)equivalence_checker.check_eqv (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_check_eqv_counterexample where [code del]:
  "norm_check_eqv_counterexample \<equiv> \<lambda>\<Sigma>. (*efficient_*)equivalence_checker.check_eqv_counterexample (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"

lemmas norm_defs = wf_rexp_def
  norm_check_eqv_def norm_closure_def norm_step_def norm_check_eqv_counterexample_def
  norm_check_eqv'_def norm_closure'_def norm_step'_def 

interpretation norm: (*efficient_*)equivalence_checker "\<sigma> \<Sigma>" \<pi> "\<epsilon> \<Sigma>" norm "\<LL> \<Sigma>"
  where "norm.check_eqv' = norm_check_eqv' \<Sigma>"
  and   "norm.check_eqv = norm_check_eqv \<Sigma>"
  and   "norm.check_eqv_counterexample = norm_check_eqv_counterexample \<Sigma>"
  and   "norm.closure' = norm_closure' \<Sigma>"
  and   "norm.closure = norm_closure \<Sigma>"
  and   "norm.step' = norm_step' \<Sigma>"
  and   "norm.step = norm_step \<Sigma>"
by unfold_locales (auto simp: norm_defs trans[OF lang_norm[OF iffD2[OF ACI_norm_wf]] ACI_norm_lang])

interpretation MSO where "alphabet.wf (set o \<sigma> \<Sigma>) = wf_rexp \<Sigma>"
by (unfold_locales) (auto simp: \<sigma>_def \<pi>_def wf_rexp_def)

definition check_eqv where
"check_eqv \<Sigma> n \<phi> \<psi> \<longleftrightarrow> wf_formula \<Sigma> n (Or \<phi> \<psi>) \<and>
   norm_check_eqv' \<Sigma> n (Plus (rexp_of \<Sigma> n (norm \<phi>)) One) (Plus (rexp_of \<Sigma> n (norm \<psi>)) One)"

definition check_eqv_counterexample where
"check_eqv_counterexample \<Sigma> n \<phi> \<psi> =
   norm_check_eqv_counterexample \<Sigma> n (Plus (rexp_of \<Sigma> n (norm \<phi>)) One) (Plus (rexp_of \<Sigma> n (norm \<psi>)) One)"

definition check_eqv' where
"check_eqv' \<Sigma> n \<phi> \<psi> \<longleftrightarrow> wf_formula \<Sigma> n (Or \<phi> \<psi>) \<and>
   norm_check_eqv' \<Sigma> n (Plus (rexp_of' \<Sigma> n (norm \<phi>)) One) (Plus (rexp_of' \<Sigma> n (norm \<psi>)) One)"
(*
definition check_eqv'' where
"check_eqv'' \<Sigma> n \<phi> \<psi> \<longleftrightarrow> wf_formula \<Sigma> n (Or \<phi> \<psi>) \<and>
   norm_check_eqv \<Sigma> n (Plus (rexp_of'' \<Sigma> n (norm \<phi>)) One) (Plus (rexp_of'' \<Sigma> n (norm \<psi>)) One)"
*)
lemma lang_Plus_Zero: "\<LL> \<Sigma> n (Plus r One) = \<LL> \<Sigma> n (Plus s One) \<longleftrightarrow> \<LL> \<Sigma> n r - {[]} = \<LL> \<Sigma> n s - {[]}"
by auto

lemmas lang\<^isub>M\<^isub>S\<^isub>O_rexp_of_norm = trans[OF sym[OF lang\<^isub>M\<^isub>S\<^isub>O_norm] lang\<^isub>M\<^isub>S\<^isub>O_rexp_of]

lemma soundness: "check_eqv \<Sigma> n \<phi> \<psi> \<Longrightarrow> lang\<^isub>M\<^isub>S\<^isub>O \<Sigma> n \<phi> = lang\<^isub>M\<^isub>S\<^isub>O \<Sigma> n \<psi>"
by (rule box_equals[OF iffD1[OF lang_Plus_Zero, OF norm.soundness']
  sym[OF trans[OF lang\<^isub>M\<^isub>S\<^isub>O_rexp_of_norm]] sym[OF trans[OF lang\<^isub>M\<^isub>S\<^isub>O_rexp_of_norm]]])
  (auto simp: check_eqv_def split: sum.splits option.splits)

lemmas lang\<^isub>M\<^isub>S\<^isub>O_rexp_of'_norm = trans[OF sym[OF lang\<^isub>M\<^isub>S\<^isub>O_norm] lang\<^isub>M\<^isub>S\<^isub>O_rexp_of']

lemma soundness': "check_eqv' \<Sigma> n \<phi> \<psi> \<Longrightarrow> lang\<^isub>M\<^isub>S\<^isub>O \<Sigma> n \<phi> = lang\<^isub>M\<^isub>S\<^isub>O \<Sigma> n \<psi>"
by (rule box_equals[OF iffD1[OF lang_Plus_Zero, OF norm.soundness']
  sym[OF trans[OF lang\<^isub>M\<^isub>S\<^isub>O_rexp_of'_norm]] sym[OF trans[OF lang\<^isub>M\<^isub>S\<^isub>O_rexp_of'_norm]]])
  (auto simp: check_eqv'_def split: sum.splits option.splits)

lemma completeness:
assumes "lang\<^isub>M\<^isub>S\<^isub>O \<Sigma> n \<phi> = lang\<^isub>M\<^isub>S\<^isub>O \<Sigma> n \<psi>" "wf_formula \<Sigma> n (Or \<phi> \<psi>)"
shows "check_eqv \<Sigma> n \<phi> \<psi>"
using assms(2) unfolding check_eqv_def
by (intro conjI[OF assms(2) norm.completeness'[OF iffD2[OF lang_Plus_Zero]],
                OF box_equals[OF assms(1) lang\<^isub>M\<^isub>S\<^isub>O_rexp_of_norm lang\<^isub>M\<^isub>S\<^isub>O_rexp_of_norm]])
   (auto simp: wf_rexp_def[symmetric] split: sum.splits option.splits intro!: wf_rexp_of)

lemma completeness':
assumes "lang\<^isub>M\<^isub>S\<^isub>O \<Sigma> n \<phi> = lang\<^isub>M\<^isub>S\<^isub>O \<Sigma> n \<psi>" "wf_formula \<Sigma> n (Or \<phi> \<psi>)"
shows "check_eqv' \<Sigma> n \<phi> \<psi>"
using assms(2) unfolding check_eqv'_def
by (intro conjI[OF assms(2) norm.completeness'[OF iffD2[OF lang_Plus_Zero]],
                OF box_equals[OF assms(1) lang\<^isub>M\<^isub>S\<^isub>O_rexp_of'_norm lang\<^isub>M\<^isub>S\<^isub>O_rexp_of'_norm]])
   (auto simp: wf_rexp_def[symmetric] split: sum.splits option.splits intro!: wf_rexp_of')

end
