(* Author: Dmitriy Traytel *)

header {* Normalization of MSO Formulas *}

(*<*)
theory MSO_Normalization
imports MSO
begin
(*>*)

context MSO
begin

fun nNot where
  "nNot (Not \<phi>) = \<phi>"
| "nNot (And \<phi>1 \<phi>2) = Or (nNot \<phi>1) (nNot \<phi>2)"
| "nNot (Or \<phi>1 \<phi>2) = And (nNot \<phi>1) (nNot \<phi>2)"
| "nNot \<phi> = Not \<phi>"

primrec norm where
  "norm (Q a m) = Q a m"
| "norm (Less m n) = Less m n"
| "norm (In m M) = In m M"
| "norm (Or \<phi> \<psi>) = Or (norm \<phi>) (norm \<psi>)"
| "norm (And \<phi> \<psi>) = And (norm \<phi>) (norm \<psi>)"
| "norm (Not \<phi>) = nNot (norm \<phi>)"
| "norm (Exists \<phi>) = Exists (norm \<phi>)"
| "norm (EXISTS \<phi>) = EXISTS (norm \<phi>)"

lemma satisfies_nNot[simp]: "satisfies (w, I) (nNot \<phi>) = satisfies (w,I) (Not \<phi>)"
by (induct \<phi> rule: nNot.induct) auto

lemma FOV_nNot[simp]: "FOV (nNot \<phi>) = FOV (Not \<phi>)"
by (induct \<phi> rule: nNot.induct) auto

lemma SOV_nNot[simp]: "SOV (nNot \<phi>) = SOV (Not \<phi>)"
by (induct \<phi> rule: nNot.induct) auto

lemma pre_wf_formula_nNot[simp]: "pre_wf_formula n (nNot \<phi>) = pre_wf_formula n (Not \<phi>)"
by (induct \<phi> rule: nNot.induct) auto

lemma FOV_norm[simp]: "FOV (norm \<phi>) = FOV \<phi>"
by (induct \<phi>) auto

lemma SOV_norm[simp]: "SOV (norm \<phi>) = SOV \<phi>"
by (induct \<phi>) auto

lemma pre_wf_formula_norm[simp]: "pre_wf_formula n (norm \<phi>) = pre_wf_formula n \<phi>"
by (induct \<phi> arbitrary: n) auto

lemma satisfies_norm[simp]: "satisfies (w, I) (norm \<phi>) = satisfies (w, I) \<phi>"
by (induct \<phi> arbitrary: I) auto

lemma lang\<^isub>M\<^isub>S\<^isub>O_norm[simp]: "lang\<^isub>M\<^isub>S\<^isub>O n (norm \<phi>) = lang\<^isub>M\<^isub>S\<^isub>O n \<phi>"
unfolding lang\<^isub>M\<^isub>S\<^isub>O_def by auto

end

end
