(* Author: Dmitriy Traytel *)

header {* WS1S *}

(*<*)
theory WS1S
imports Formulas Regular_Operators "~~/src/HOL/BNF/Examples/Stream"
begin
(*>*)

subsection {* Encodings *}

definition "cutSame x s = stake (LEAST n. sdrop n s = same x) s"

abbreviation "poss I \<equiv> (\<Union>x\<in>set I. case x of Inl p \<Rightarrow> {p} | Inr P \<Rightarrow> P)"

lemma shift_snth: "(xs @- s) !! n = (if n < length xs then xs ! n else s !! (n - length xs))"
  by auto

lemma stream_map_stream_map2[simp]:
  "stream_map f (stream_map2 g s1 s2) = stream_map2 (\<lambda>x y. f (g x y)) s1 s2"
  unfolding stream_map2_szip stream.map_comp' o_def split_def ..

lemma stream_map2_alt:
  "(stream_map2 f s1 s2 = s) = (\<forall>n. f (s1 !! n) (s2 !! n) = s !! n)"
  unfolding stream_map2_szip stream_map_alt by auto

lemma snth_stream_map2[simp]:
  "stream_map2 f s1 s2 !! n = f (s1 !! n) (s2 !! n)"
  by (induct n arbitrary: s1 s2) auto

lemma stake_stream_map2[simp]:
  "stake n (stream_map2 f s1 s2) = map (split f) (zip (stake n s1) (stake n s2))"
  by (induct n arbitrary: s1 s2) auto

lemma sdrop_stream_map2[simp]:
  "sdrop n (stream_map2 f s1 s2) = stream_map2 f (sdrop n s1) (sdrop n s2)"
  by (induct n arbitrary: s1 s2) auto

lemma stake_szip[simp]:
  "stake n (szip s1 s2) = zip (stake n s1) (stake n s2)"
  by (induct n arbitrary: s1 s2) auto

lemma sdrop_szip[simp]: "sdrop n (szip s1 s2) = szip (sdrop n s1) (sdrop n s2)"
  by (induct n arbitrary: s1 s2) auto

lemma take_stake: "take n (stake m s) = stake (min n m) s"
proof (induct m arbitrary: s n)
  case (Suc m) thus ?case by (cases n) auto
qed simp

lemma drop_stake: "drop n (stake m s) = stake (m - n) (sdrop n s)"
proof (induct m arbitrary: s n)
  case (Suc m) thus ?case by (cases n) auto
qed simp

lemma stream_map_same[simp]: "stream_map f (same x) = same (f x)"
  by (coinduct rule: stream.coinduct[of "\<lambda>s1 s2. s1 = stream_map f (same x) \<and> s2 = same (f x)"]) auto

lemma (in wellorder) min_Least:
  "\<lbrakk>\<exists>n. P n; \<exists>n. Q n\<rbrakk> \<Longrightarrow> min (Least P) (Least Q) = (LEAST n. P n \<or> Q n)"
proof (intro sym[OF Least_equality])
  fix y assume "P y \<or> Q y"
  thus "min (Least P) (Least Q) \<le> y"
  proof (elim disjE)
    assume "P y"
    hence "Least P \<le> y" by (auto intro: LeastI2_wellorder)
    thus "min (Least P) (Least Q) \<le> y" unfolding min_def by auto
  next
    assume "Q y"
    hence "Least Q \<le> y" by (auto intro: LeastI2_wellorder)
    thus "min (Least P) (Least Q) \<le> y" unfolding min_def by auto
  qed
qed (metis LeastI_ex min_def)

lemma sdrop_snth: "sdrop n s !! m = s !! (n + m)"
  by (induct n arbitrary: m s) auto

context formula
begin

definition "any \<equiv> hd \<Sigma>"

lemma any_\<Sigma>[simp]: "any \<in> set \<Sigma>"
  unfolding any_def by (auto simp: nonempty intro: someI[of _ "hd \<Sigma>"])

lemma enc_atom_any_\<sigma>[simp]: "length I = n \<Longrightarrow> enc_atom I m any \<in> set (\<sigma> \<Sigma> n)"
  by (auto simp: \<sigma>_def image_iff set_n_lists)

fun stream_enc :: "'a interp \<Rightarrow> ('a \<times> bool list) stream" where
  "stream_enc (w, I) = stream_map2 (enc_atom I) nats (w @- same any)"

lemma tl_stream_enc[simp]: "stream_map \<pi> (stream_enc (w, x # I)) = stream_enc (w, I)"
  by (auto simp: comp_def \<pi>_def)

lemma enc_atom_max: "\<lbrakk>\<forall>x\<in>set I. case x of Inl p \<Rightarrow> p \<le> n | Inr P \<Rightarrow> \<forall>p\<in>P. p \<le> n; n \<le> n'\<rbrakk> \<Longrightarrow>
  enc_atom I (Suc n') a = (a, replicate (length I) False)"
  by (induct I) (auto split: sum.splits)

lemma ex_Loop_stream_enc:
assumes "\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True"
shows "\<exists>n. sdrop n (stream_enc (w, I)) = same (any, replicate (length I) False)"
proof -
  from assms have "\<exists>n > length w. \<forall>x\<in>set I. case x of Inl p \<Rightarrow> p \<le> n | Inr P \<Rightarrow> \<forall>p\<in>P. p \<le> n"
  proof (induct I)
    case (Cons x I)
    then obtain n where IH: "length w < n"
      "\<forall>x\<in>set I. case x of Inl p \<Rightarrow> p \<le> n | Inr P \<Rightarrow> \<forall>p\<in>P. p \<le> n" by auto
    thus ?case
    proof (cases x)
      case (Inl p)
      with IH show ?thesis
        by (intro exI[of _ "max p n"]) (fastforce split: sum.splits)
    next
      case (Inr P)
      with IH Cons(2) show ?thesis
        by (intro exI[of _ "max (Max P) n"]) (fastforce dest: Max_ge split: sum.splits)
    qed
  qed auto
  then guess n by (elim exE conjE)
  hence "sdrop (Suc n) (stream_enc (w, I)) = same (any, replicate (length I) False)"
    (is "?s1 n = ?s2")
    by (coinduct rule: stream.coinduct[of "\<lambda>s1 s2. \<exists>n'\<ge> n. s1 = ?s1 n' \<and> s2 = ?s2"])
      (auto simp: enc_atom_max dest: le_SucI)
  thus ?thesis by blast
qed

lemma length_snth_enc[simp]: "length (snd (stream_enc (w, I) !! n)) = length I"
  by auto

lemma stream_set_same: "\<lbrakk>y \<in> stream_set s; s = same x\<rbrakk> \<Longrightarrow> y = x"
  by (induct rule: stream_set_induct1) auto

lemma same_alt: "s = same x \<longleftrightarrow> stream_set s = {x}"
  using stream_set_same[of _ "same x" x]
  apply auto
  apply (metis shd_stream_set)
  apply (coinduct rule: stream.coinduct[of "\<lambda>s1 s2. s2 = same x \<and> stream_set s1 = {x}"])
  apply auto
  apply (metis shd_stream_set singleton_iff)
  apply (metis stl_stream_set singleton_iff)
  by (metis (full_types) empty_iff insert_iff shd_stream_set stl_stream_set)

lemma sdrop_sameE: "\<lbrakk>sdrop n (w @- same y) = same y; p < length w; \<not> p < n\<rbrakk> \<Longrightarrow> w ! p = y"
  unfolding not_less same_alt
  apply (induct p arbitrary: w n)
  apply simp
  apply (metis hd_conv_nth shd_stream_set shift_simps(1) singletonE stream_set_shift)
  apply (case_tac w)
  apply simp_all
  apply (case_tac n)
  apply simp_all
  by (metis equals0D nth_mem singletonE subset_singletonD)

lemma less_length_cutSame:
  "\<lbrakk>(w @- same y) !! p = a\<rbrakk> \<Longrightarrow> a = y \<or> (p < length (cutSame y (w @- same y)) \<and> w ! p = a)"
  unfolding cutSame_def length_stake
  by (rule LeastI2_ex[OF exI[of _ "length w"]])
   (auto simp: sdrop_shift shift_snth split: split_if_asm elim: sdrop_sameE)

lemma less_length_cutSame_Inl:
  "\<lbrakk>(\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True); r < length I; I ! r = Inl p\<rbrakk> \<Longrightarrow>
  p < length (cutSame (any, replicate (length I) False) (stream_enc (w, I)))"
  unfolding cutSame_def length_stake
  by (erule LeastI2_ex[OF ex_Loop_stream_enc ccontr])
   (auto simp: stream_map2_alt dest!: add_diff_inverse,
   metis (lifting, full_types) nth_map nth_replicate sum.simps(5))

lemma less_length_cutSame_Inr:
  "\<lbrakk>(\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True); r < length I; I ! r = Inr P\<rbrakk> \<Longrightarrow>
  \<forall>p \<in> P. p < length (cutSame (any, replicate (length I) False) (stream_enc (w, I)))"
  unfolding cutSame_def length_stake
  by (rule ballI, erule LeastI2_ex[OF ex_Loop_stream_enc ccontr])
    (auto simp: stream_map2_alt dest!: add_diff_inverse,
      metis nth_map nth_replicate sum.simps(6))

fun enc :: "'a interp \<Rightarrow> ('a \<times> bool list) list set" where
  "enc (w, I) = {x. \<exists>n. x = (cutSame (any, replicate (length I) False) (stream_enc (w, I)) @
     replicate n (any, replicate (length I) False))}"

lemma cutSame_all[simp]: "cutSame x (same x) = []"
  unfolding cutSame_def by (auto intro: Least_equality)

lemma cutSame_stop[simp]:
  assumes "x \<noteq> y"
  shows "cutSame x (xs @- Stream y (same x)) = xs @ [y]" (is "cutSame x ?s = _")
proof -
  have "(LEAST n. sdrop n ?s = same x) = Suc (length xs)"
  proof (rule Least_equality)
    show "sdrop (Suc (length xs)) ?s = same x"
      by (metis sdrop_shift sdrop_simps(2) stream.sels(2))
  next
    fix m assume *: "sdrop m ?s = same x"
    { assume "m < Suc (length xs)"
      hence "m \<le> length xs" by simp
      then obtain ys where "sdrop m ?s = ys @- Stream y (same x)"
        by atomize_elim (induct m arbitrary: xs, auto)
      with * obtain "ys @- Stream y (same x) = same x" by simp
      hence "Stream y (same x) = same x" by (metis sdrop_same sdrop_shift)
      with assms have False by (metis same_simps(1) stream.sels(1))
    }
    thus "Suc (length xs) \<le> m" by (blast intro: leI)
  qed
  thus ?thesis unfolding cutSame_def
    by (metis length_append_singleton shift.simps shift_append stake_shift)
qed

lemma cutSame_shift_same: "\<exists>n. w = cutSame x (w @- same x) @ replicate n x"
proof (induct w rule: rev_induct)
  case (snoc a w)
  then obtain n where "w = cutSame x (w @- same x) @ replicate n x" by blast
  thus ?case
    by (cases "a = x")
     (auto simp: same_unfold[symmetric] replicate_append_same[symmetric] intro!: exI[of _ "Suc n"])
qed simp

lemma set_cutSame: "set (cutSame x (w @- same x)) \<subseteq> set w"
proof (induct w rule: rev_induct)
  case (snoc a w)
  thus ?case by (cases "a = x") (auto simp: same_unfold[symmetric])
qed simp

lemma stream_enc_cutSame:
  assumes "(\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True)"
  shows "stream_enc (w, I) = cutSame (any, replicate (length I) False) (stream_enc (w, I)) @-
    same (any, replicate (length I) False)"
  unfolding cutSame_def
  by (rule trans[OF sym[OF stake_sdrop] arg_cong2[of _ _ _ _ "op @-", OF refl]])
     (rule LeastI_ex[OF ex_Loop_stream_enc[OF assms]])

lemma map_fst_zip_min[simp]: "map fst (zip xs ys) \<equiv> take (min (length xs) (length ys)) xs"
proof (induct ys arbitrary: xs)
  case Cons thus ?case by (case_tac xs) auto
qed simp

lemma map_snd_zip_min[simp]: "map snd (zip xs ys) \<equiv> take (min (length xs) (length ys)) ys"
proof (induct ys arbitrary: xs)
  case Cons thus ?case by (case_tac xs) auto
qed simp

lemma stream_enc_enc:
  assumes "(\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True)" and  v: "v \<in> enc (w, I)"
  shows "stream_enc (w, I) = v @- same (any, replicate (length I) False)"
  (is "?s = ?v @- same ?F")
proof -
  from assms(1) obtain n where "sdrop n (stream_enc (w, I)) = same ?F" by (metis ex_Loop_stream_enc)
  moreover from v obtain m where "?v = cutSame ?F ?s @ replicate m ?F" by auto
  ultimately show "?s = v @- same ?F"
    by (auto simp del: stream_enc.simps intro: stream_enc_cutSame[OF assms(1)])
qed

lemma stream_enc_enc_some:
  assumes "(\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True)"
  shows "stream_enc (w, I) = (SOME v. v \<in> enc (w, I)) @- same (any, replicate (length I) False)"
  by (rule stream_enc_enc[OF assms], rule someI_ex) auto

lemma enc_unique_length: "v \<in> enc (w, I) \<Longrightarrow> \<forall>v'. length v' = length v \<and> v' \<in> enc (w, I) \<longrightarrow> v = v'"
  by auto

lemma sdrop_same: "sdrop n s = same x \<Longrightarrow> n \<le> m \<Longrightarrow> s !! m = x"
  by (metis le_iff_add sdrop_snth snth_same)

lemma fin_cutSame_tl:
  assumes "\<exists>n. sdrop n s = same x" 
  shows "fin_cutSame (\<pi> x) (map \<pi> (cutSame x s)) = cutSame (\<pi> x) (stream_map \<pi> s)"
proof -
  def min \<equiv> "LEAST n. sdrop n s = same x"
  from assms have min: "sdrop min s = same x" "\<And>m. sdrop m s = same x \<Longrightarrow> min \<le> m"
    unfolding min_def by (auto intro: LeastI Least_le)
  have Ex: "\<exists>n. drop n (map \<pi> (stake min s)) = replicate (length (map \<pi> (stake min s)) - n) (\<pi> x)"
    by (auto intro: exI[of _ "length (map \<pi> (stake min s))"])
  have "fin_cutSame (\<pi> x) (map \<pi> (cutSame x s)) =
     map \<pi> (stake  (LEAST n.
       map \<pi> (stake (min - n) (sdrop n s)) = replicate (min - n) (\<pi> x) \<or> sdrop n s = same x) s)"
    unfolding fin_cutSame_def cutSame_def take_map take_stake min_Least[OF Ex assms, folded min_def]
      min_def[symmetric] by (auto simp: drop_map drop_stake)
  also have "(\<lambda>n. map \<pi> (stake (min - n) (sdrop n s)) = replicate (min - n) (\<pi> x) \<or> sdrop n s = same x) =
     (\<lambda>n. stream_map \<pi> (sdrop n s) = same (\<pi> x))"
  proof (rule ext, unfold stream_map_alt snth_same, safe)
    fix n m
    assume "map \<pi> (stake (min - n) (sdrop n s)) = replicate (min - n) (\<pi> x)"
    hence "\<forall>y\<in>set (stake (min - n) (sdrop n s)). \<pi> y = \<pi> x"
      by (intro iffD1[OF map_eq_conv]) (metis length_stake map_replicate_const)
    hence "\<forall>i<min - n. \<pi> (sdrop n s !! i) = \<pi> x"
      unfolding all_set_conv_all_nth by (auto simp: sdrop_snth)
    thus "\<pi> (sdrop n s !! m) = \<pi> x"
    proof (cases "m < min - n")
      case False
      hence "min \<le> n + m" by linarith
      hence "sdrop n s !! m = x" unfolding sdrop_snth by (rule sdrop_same[OF min(1)])
      thus ?thesis by simp
    qed auto
  next
    fix n
    assume "\<forall>m. \<pi> (sdrop n s !! m) = \<pi> x"
    thus "map \<pi> (stake (min - n) (sdrop n s)) = replicate (min - n) (\<pi> x)"
      unfolding stake_stream_map[symmetric] by (metis snth_same stake_same stream_map_alt)
  qed auto
  finally show ?thesis unfolding cutSame_def sdrop_stream_map stake_stream_map .
qed

lemma tl_enc[simp]:
  assumes "\<forall>x \<in> set (x # I). case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True"
  shows "SAMEQUOT (any, replicate (length I) False) (map \<pi> ` enc (w, x # I)) = enc (w, I)"
  unfolding SAMEQUOT_def
  by (fastforce simp: assms \<pi>_def
    fin_cutSame_tl[OF ex_Loop_stream_enc[OF assms], unfolded \<pi>_def, simplified, symmetric])

lemma encD:
  "\<lbrakk>v \<in> enc (w, I); (\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True)\<rbrakk> \<Longrightarrow>
  v = map (split (enc_atom I)) (zip [0 ..< length v] (stake (length v) (w @- same any)))"
  by (erule box_equals[OF sym[OF arg_cong[of _ _ "stake (length v)" ,OF stream_enc_enc]]])
   (auto simp: stake_shift sdrop_shift stake_add[symmetric] simp del: stake_add)

lemma enc_Inl: "\<lbrakk>x \<in> enc (w, I); (\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True);
  m < length I; I ! m = Inl p\<rbrakk> \<Longrightarrow> p < length x \<and> snd (x ! p) ! m"
  by (auto dest!: less_length_cutSame_Inl[of _ _ _ w] simp: nth_append cutSame_def)

lemma enc_Inr: assumes "x \<in> enc (w, I)" "\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True"
  "M < length I" "I ! M = Inr P"
  shows "p \<in> P \<longleftrightarrow> p < length x \<and> snd (x ! p) ! M"
proof
  assume "p \<in> P" with assms show "p < length x \<and> snd (x ! p) ! M"
    by (auto dest!: less_length_cutSame_Inr[of _ _ _ w] simp: nth_append cutSame_def)
next
  assume "p < length x \<and> snd (x ! p) ! M"
  thus "p \<in> P" using assms by (subst (asm) (2) encD[OF assms(1,2)]) auto
qed

lemma enc_length:
  assumes "enc (w, I) = enc (w', I')"
  shows "length I = length I'"
proof -
  let ?cL = "\<lambda>w I. cutSame (any, replicate (length I) False) (stream_enc (w, I))"
  let ?w = "\<lambda>w I m. ?cL w I @ replicate (m - length (?cL w I)) (any, replicate (length I) False)"
  let ?max = "max (length (?cL w I)) (length (?cL w' I')) + 1"
  from assms have "?w w I ?max \<in> enc (w, I)" "?w w' I' ?max \<in> enc (w', I')" by auto
  hence "?w w I ?max = ?w w' I' ?max" using enc_unique_length assms by (simp del: enc.simps)
  moreover have "last (?w w I ?max) = (any, replicate (length I) False)"
                "last (?w w' I' ?max) = (any, replicate (length I') False)" by auto
  ultimately show "length I = length I'" by auto
qed

lemma enc_stream_enc: 
  "\<lbrakk>(\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True);
    (\<forall>x \<in> set I'. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True); 
    enc (w, I) = enc (w', I')\<rbrakk> \<Longrightarrow> stream_enc (w, I) = stream_enc (w', I')"
  by (rule box_equals[OF _ sym[OF stream_enc_enc_some] sym[OF stream_enc_enc_some]])
    (auto dest: enc_length simp del: enc.simps)

fun wf_interp_for_formula :: "'a interp \<Rightarrow> 'a formula \<Rightarrow> bool" where
  "wf_interp_for_formula (w, I) \<phi> =
    ((\<forall>a \<in> set w. a \<in> set \<Sigma>) \<and>
     (\<forall>n \<in> FOV \<phi>. case I ! n of Inl _ \<Rightarrow> True | _ \<Rightarrow> False) \<and>
     (\<forall>n \<in> SOV \<phi>. case I ! n of Inl _ \<Rightarrow> False | Inr _ \<Rightarrow> True) \<and>
     (\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True))"

fun satisfies :: "'a interp \<Rightarrow> 'a formula \<Rightarrow> bool" (infix "\<Turnstile>" 50) where
  "(w, I) \<Turnstile> FQ a m = ((case I ! m of Inl p \<Rightarrow> if p < length w then w ! p else any) = a)"
| "(w, I) \<Turnstile> FLess m1 m2 = ((case I ! m1 of Inl p \<Rightarrow> p) < (case I ! m2 of Inl p \<Rightarrow> p))"
| "(w, I) \<Turnstile> FIn m M = ((case I ! m of Inl p \<Rightarrow> p) \<in> (case I ! M of Inr P \<Rightarrow> P))"
| "(w, I) \<Turnstile> FNot \<phi> = (\<not> (w, I) \<Turnstile> \<phi>)"
| "(w, I) \<Turnstile> (FOr \<phi>\<^isub>1 \<phi>\<^isub>2) = ((w, I) \<Turnstile> \<phi>\<^isub>1 \<or> (w, I) \<Turnstile> \<phi>\<^isub>2)"
| "(w, I) \<Turnstile> (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2) = ((w, I) \<Turnstile> \<phi>\<^isub>1 \<and> (w, I) \<Turnstile> \<phi>\<^isub>2)"
| "(w, I) \<Turnstile> (FExists \<phi>) = (\<exists>p. (w, Inl p # I) \<Turnstile> \<phi>)"
| "(w, I) \<Turnstile> (FEXISTS \<phi>) = (\<exists>P. finite P \<and> (w, Inr P # I) \<Turnstile> \<phi>)"

definition lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S :: "nat \<Rightarrow> 'a formula \<Rightarrow> ('a \<times> bool list) list set" where
  "lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi> = \<Union>{enc (w, I) | w I . length I = n \<and> wf_interp_for_formula (w, I) \<phi> \<and> (w, I) \<Turnstile> \<phi>}"

lemma encD_ex: "\<lbrakk>x \<in> enc (w, I); (\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True)\<rbrakk> \<Longrightarrow>
  \<exists>n. x = map (split (enc_atom I)) (zip [0 ..< n] (stake n (w @- same any)))"
  by (auto dest!: encD simp del: enc.simps)

lemma enc_set_\<sigma>: "\<lbrakk>x \<in> enc (w, I); (\<forall>x \<in> set I. case x of Inr P \<Rightarrow> finite P | _ \<Rightarrow> True);
  length I = n; a \<in> set x; set w \<subseteq> set \<Sigma>\<rbrakk> \<Longrightarrow> a \<in> set (\<sigma> \<Sigma> n)"
  apply (auto dest!: encD_ex simp: in_set_zip simp del: enc.simps)
  apply (case_tac "na < length w")
  apply (auto intro!: enc_atom_\<sigma>)
  done

definition "positions_in_row s i =
  Option.these (stream_set (stream_map2 (\<lambda>p (_, bs). if nth bs i then Some p else None) nats s))"

lemma positions_in_row: "positions_in_row s i = {p. snd (s !! p) ! i}"
  unfolding positions_in_row_def these_def stream_map2_szip stream.set_natural' stream_set_range
  by (auto split: split_if_asm intro!: image_eqI[of _ the] split: prod.splits)

lemma positions_in_row_unique: "\<exists>!p. snd (s !! p) ! i \<Longrightarrow>
  the_elem (positions_in_row s i) = (THE p. snd (s !! p) ! i)"
  by (rule the1I2) (auto simp: the_elem_def positions_in_row)

lemma positions_in_row_nth: "\<exists>!p. snd (s !! p) ! i \<Longrightarrow>
  snd (s !! the_elem (positions_in_row s i)) ! i"
  unfolding positions_in_row_unique by (rule the1I2) auto

definition "dec_word s = cutSame any (stream_map fst s)"

lemma dec_word_stream_enc: "dec_word (stream_enc (w, I)) = cutSame any (w @- same any)"
  unfolding dec_word_def by (auto intro!: arg_cong[of _ _ "cutSame any"] simp: stream_map2_alt)

definition "stream_dec n FO s = map (\<lambda>i.
  if i \<in> FO
  then Inl (the_elem (positions_in_row s i))
  else Inr (positions_in_row s i)) [0..<n]"

lemma stream_dec_Inl: "\<lbrakk>i \<in> FO; i < n\<rbrakk> \<Longrightarrow> \<exists>p. stream_dec n FO s ! i = Inl p"
  unfolding stream_dec_def using nth_map[of n "[0..<n]"] by auto

lemma stream_dec_not_Inr: "\<lbrakk>stream_dec n FO s ! i = Inr P; i \<in> FO; i < n\<rbrakk> \<Longrightarrow> False"
  unfolding stream_dec_def using nth_map[of n "[0..<n]"] by auto

lemma stream_dec_Inr: "\<lbrakk>i \<notin> FO; i < n\<rbrakk> \<Longrightarrow> \<exists>P. stream_dec n FO s ! i = Inr P"
  unfolding stream_dec_def using nth_map[of n "[0..<n]"] by auto

lemma stream_dec_not_Inl: "\<lbrakk>stream_dec n FO s ! i = Inl p; i \<notin> FO; i < n\<rbrakk> \<Longrightarrow> False"
  unfolding stream_dec_def using nth_map[of n "[0..<n]"] by auto

lemma Inr_dec_finite: "\<lbrakk>\<forall>i<n. finite {p. snd (s !! p) ! i}; Inr P \<in> set (stream_dec n FO s)\<rbrakk> \<Longrightarrow>
  finite P"
  unfolding stream_dec_def by (auto simp: positions_in_row)

lemma enc_atom_dec: 
  "\<lbrakk>\<forall>p. length (snd (s !! p)) = n; \<forall>i \<in> FO. i < n \<longrightarrow> (\<exists>!p. snd (s !! p) ! i); a = fst (s !! p)\<rbrakk> \<Longrightarrow>
   enc_atom (stream_dec n FO s) p a = s !! p"
  unfolding stream_dec_def
  by (rule sym, subst surjective_pairing[of "s !! p"])
    (auto intro!: nth_equalityI simp: positions_in_row simp del: pair_collapse split: split_if_asm,
    (metis positions_in_row positions_in_row_nth)+)

lemma length_stream_dec[simp]: "length (stream_dec n FO x) = n"
  unfolding stream_dec_def by auto

lemma stream_enc_dec:
  "\<lbrakk>\<exists>n. sdrop n (stream_map fst s) = same any;
   stream_all (\<lambda>x. length (snd x) = n) s; \<forall>i \<in> FO. (\<exists>!p. snd (s !! p) ! i)\<rbrakk> \<Longrightarrow>
   stream_enc (dec_word s, stream_dec n FO s) = s"
  unfolding dec_word_def snth_fromN
  by (drule LeastI_ex)
    (auto intro!: enc_atom_dec simp: stream_map2_alt cutSame_def
    simp del: stake_stream_map sdrop_stream_map
    intro!: trans[OF arg_cong2[of _ _ _ _ "op !!"] snth_stream_map]
            trans[OF arg_cong2[of _ _ _ _ "op @-"] stake_sdrop])

lemma stream_enc_unique: 
  "i < length I \<Longrightarrow> \<exists>p. I ! i = Inl p \<Longrightarrow> \<exists>!p. snd (stream_enc (w, I) !! p) ! i"
  by auto

lemma stream_dec_enc_Inl:
  "\<lbrakk>stream_dec n FO (stream_enc (w, I)) ! i = Inl p'; I ! i = Inl p; i \<in> FO; i < n; length I = n\<rbrakk> \<Longrightarrow>
  p = p'"
  unfolding stream_dec_def
  by (auto intro!: trans[OF _ sym[OF  positions_in_row_unique[OF stream_enc_unique]]]
    simp del: stream_enc.simps) simp

lemma stream_dec_enc_Inr:
  "\<lbrakk>stream_dec n FO (stream_enc (w, I)) ! i = Inr P'; I ! i = Inr P; i \<notin> FO; i < n; length I = n\<rbrakk> \<Longrightarrow>
  P = P'"
  unfolding stream_dec_def positions_in_row by auto

lemma Collect_snth: "{p. P (Stream x s !! p)} \<subseteq> {0} \<union> Suc ` {p. P (s !! p)}"
  unfolding image_def by (auto simp: gr0_conv_Suc)

lemma finite_True_in_row: "\<forall>i < n. finite {p. snd ((w @- same (any, replicate n False)) !! p) ! i}"
  by (induct w) (auto intro: finite_subset[OF Collect_snth])

lemma lang_ENC:
 assumes "wf_formula n \<phi>"
 shows "lang n (ENC n \<phi>) = \<Union>{enc (w, I) | w I . length I = n \<and> wf_interp_for_formula (w, I) \<phi>}"
   (is "?L = ?R")
proof (intro equalityI subsetI)
  fix x assume L: "x \<in> ?L"
  hence *: "set x \<subseteq> set (\<sigma> \<Sigma> n)" using wf_lang_wf_word[OF wf_rexp_ENC] by (auto simp: wf_word)
  let ?s = "x @- same (any, replicate n False)"
  have "list_all (\<lambda>bs. length (snd bs) = n) x"
    using bspec[OF wf_lang_wf_word[OF wf_rexp_ENC], OF `x \<in> ?L`]
    by (auto simp: list_all_iff wf_word) (auto simp: \<sigma>_def set_n_lists)
  hence "stream_all (\<lambda>x. length (snd x) = n) (x @- same (any, replicate n False))"
    by (auto simp only: stream_all_shift stream_all_same length_replicate snd_conv)
  moreover
  { fix m assume "m \<in> FOV \<phi>"
    with assms have "m < n" by (auto simp: max_idx_vars)
    with L `m \<in> FOV \<phi>` obtain u z v where uzv: "x = u @ z @ v"
      "u \<in> star (lang n (arbitrary_except n [(m, False)] \<Sigma>))"
      "z \<in> lang n (arbitrary_except n [(m, True)] \<Sigma>)"
      "v \<in> star (lang n (arbitrary_except n [(m, False)] \<Sigma>))" unfolding ENC_def
      by (auto simp: wf_rexp_valid_ENC finite_FOV dest!: iffD1[OF lang_flatten_INTERSECT, rotated -1])
        (fastforce simp: valid_ENC_def)
    with `m < n` have "\<exists>!p.  snd (x ! p) ! m \<and> p < length x"
    proof (intro ex1I[of _ "length u"])
      fix p assume "m < n" "snd (x ! p) ! m \<and> p < length x"
      with star_arbitrary_except[OF uzv(2)] arbitrary_except[OF uzv(3)] star_arbitrary_except[OF uzv(4)]
        show "p = length u" by (cases rule: nat_less_cases) (auto simp: nth_append uzv(1))
    qed (auto dest!: arbitrary_except)
    then obtain p where p: "p < length x" "snd (x ! p) ! m"
      "\<And>q. snd (x ! q) ! m \<and> q < length x \<longrightarrow> q = p" by auto
    hence "\<exists>!p. snd (?s !! p) ! m"
    proof (intro ex1I[of _ p])
      fix q from p `m < n` show "snd (?s !! q) ! m \<Longrightarrow> q = p" by (cases "q < length x") auto
    qed auto
  }
  moreover have "sdrop (length x) (stream_map fst (x @- same (any, replicate n False))) = same any"
    unfolding sdrop_stream_map by (subst sdrop_shift[OF refl refl]) simp
  ultimately have enc_dec: "stream_enc (dec_word ?s, stream_dec n (FOV \<phi>) ?s) =
    x @- same (any, replicate n False)" by (intro stream_enc_dec) auto
  def I \<equiv> "stream_dec n (FOV \<phi>) ?s"
  with assms have "wf_interp_for_formula (dec_word ?s, I) \<phi>" unfolding I_def dec_word_def
    by (auto dest: stream_dec_not_Inr stream_dec_not_Inl simp :\<sigma>_def  max_idx_vars
      dest!: set_mp[OF set_cutSame[of any "map fst x"]] set_mp[OF *]  split: sum.splits)
      (auto simp: stream_dec_def positions_in_row finite_True_in_row)
  moreover have "length I = n" unfolding I_def by simp
  moreover have "x \<in> enc (dec_word ?s, I)" unfolding I_def
    by (simp add: enc_dec cutSame_shift_same del: stream_enc.simps)
  ultimately show "x \<in> ?R" by blast
next
  fix x assume "x \<in> ?R"
  then obtain w I where I: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) \<phi>" "length I = n" by blast
  { fix i from I(2) have "(w @- same any) !! i \<in> set \<Sigma>" by (cases "i < length w") auto } note * = this
  from I have "x @- same (any, replicate (length I) False) = stream_enc (w, I)" (is "x @- ?F = ?s")
    by (intro stream_enc_enc[symmetric]) auto
  with * `length I = n` have "\<forall>x \<in> set x. length (snd x) = n \<and> fst x \<in> set \<Sigma>"
    by (auto dest!: shift_snth_less[of _ _ ?F, symmetric] simp: in_set_conv_nth)
  thus "x \<in> ?L"
  proof (cases "FOV \<phi> = {}")
    case False
    hence nonempty: "valid_ENC n ` FOV \<phi> \<noteq> {}" by simp
    have finite: "finite (valid_ENC n ` FOV \<phi>)" by (rule finite_imageI[OF finite_FOV])
    from False assms(1) have "0 < n" by (cases n) (auto split: dest!: max_idx_vars)
    with wf_rexp_valid_ENC have wf_rexp: "\<forall>x \<in> valid_ENC n ` FOV \<phi>. wf n x" by auto
    { fix r assume "r \<in> FOV \<phi>"
      with I(2) obtain p where p: "I ! r = Inl p" by (cases "I ! r") auto
      from `r \<in> FOV \<phi>` assms I(2,3) have r: "r < length I" by (auto dest!: max_idx_vars)
      from p I(1,2) r have "p < length x"
        using less_length_cutSame_Inl[of I r p w] by auto
      with p I r * have "[x ! p] \<in> lang n (arbitrary_except n [(r, True)] \<Sigma>)"
        by (subst encD[of x]) (auto intro!: enc_atom_lang_arbitrary_except_True)
      moreover
      from p I r * have "take p x \<in> star (lang n (arbitrary_except n [(r, False)] \<Sigma>))"
        by (subst encD[of x]) (auto simp: in_set_conv_nth intro!: Ball_starI enc_atom_lang_arbitrary_except_False)
      moreover
      from p I r * have "drop (Suc p) x \<in> star (lang n (arbitrary_except n [(r, False)] \<Sigma>))"
        by (subst encD[of x]) (auto simp: in_set_conv_nth simp del: snth.simps intro!: Ball_starI enc_atom_lang_arbitrary_except_False)
      ultimately have "take p x @ [x ! p] @ drop (p + 1) x \<in> lang n (valid_ENC n r)"
        using `0 < n` unfolding valid_ENC_def by (auto simp del: append.simps)
      hence "x \<in> lang n (valid_ENC n r)" using id_take_nth_drop[OF `p < length x`] by auto
    }
    with False lang_flatten_INTERSECT[OF finite nonempty wf_rexp] show ?thesis by (auto simp: ENC_def)
  qed (simp add: ENC_def, auto simp: \<sigma>_def set_n_lists image_iff)
qed

subsection {* Welldefinedness of enc wrt. Models *}

lemma wf_interp_for_formula_FExists: 
 "\<lbrakk>wf_formula (length I) (FExists \<phi>)\<rbrakk>\<Longrightarrow>
  wf_interp_for_formula (w, I) (FExists \<phi>) \<longleftrightarrow> (\<forall>p. wf_interp_for_formula (w, Inl p # I) \<phi>)"
  by (auto simp: nth_Cons' split: split_if_asm)

lemma wf_interp_for_formula_any_Inl: "wf_interp_for_formula (w, Inl p # I) \<phi> \<Longrightarrow>
  \<forall>p. wf_interp_for_formula (w, Inl p # I) \<phi>"
  by (auto simp: nth_Cons' split: split_if_asm)

lemma wf_interp_for_formula_FEXISTS: 
 "\<lbrakk>wf_formula (length I) (FEXISTS \<phi>)\<rbrakk>\<Longrightarrow>
  wf_interp_for_formula (w, I) (FEXISTS \<phi>) \<longleftrightarrow> (\<forall>P. finite P \<longrightarrow> wf_interp_for_formula (w, Inr P # I) \<phi>)"
  by (auto simp: nth_Cons' split: split_if_asm)

lemma wf_interp_for_formula_any_Inr: "wf_interp_for_formula (w, Inr P # I) \<phi> \<Longrightarrow>
  \<forall>P. finite P \<longrightarrow> wf_interp_for_formula (w, Inr P # I) \<phi>"
  by (auto simp: nth_Cons' split: split_if_asm)

lemma wf_interp_for_formula_FOr:
  "wf_interp_for_formula (w, I) (FOr \<phi>1 \<phi>2) =
     (wf_interp_for_formula (w, I) \<phi>1 \<and> wf_interp_for_formula (w, I) \<phi>2)"
  by auto

lemma wf_interp_for_formula_FAnd:
  "wf_interp_for_formula (w, I) (FAnd \<phi>1 \<phi>2) =
     (wf_interp_for_formula (w, I) \<phi>1 \<and> wf_interp_for_formula (w, I) \<phi>2)"
  by auto

lemma enc_wf_interp: 
 "\<lbrakk>wf_formula (length I) \<phi>; wf_interp_for_formula (w, I) \<phi>; x \<in> enc (w, I)\<rbrakk> \<Longrightarrow>
  wf_interp_for_formula (dec_word (x @- same (any, replicate (length I) False)),
    stream_dec (length I) (FOV \<phi>) (x @- same (any, replicate (length I) False))) \<phi>"
  using
    stream_dec_Inl[of _ "FOV \<phi>" "length I" "stream_enc (w, I)", OF _ bspec[OF max_idx_vars]]
    stream_dec_Inr[of _ "FOV \<phi>" "length I" "stream_enc (w, I)", OF _ bspec[OF max_idx_vars]]
  by (auto split: sum.splits intro: Inr_dec_finite[OF finite_True_in_row] simp: max_idx_vars dec_word_def
    dest!: stream_dec_not_Inl stream_dec_not_Inr set_mp[OF set_cutSame] simp del: stream_enc.simps)
    (auto simp: cutSame_def in_set_zip stream_map2_alt shift_snth)

lemma enc_atom_welldef: "\<forall>x a. enc_atom I x a = enc_atom I' x a \<Longrightarrow> m < length I \<Longrightarrow>
  (case (I ! m, I' ! m) of (Inl p, Inl q) \<Rightarrow> p = q | (Inr P, Inr Q) \<Rightarrow> P = Q | _ \<Rightarrow> True)"
proof (induct "length I" arbitrary: I I' m)
  case (Suc n I I')
  then obtain x xs x' xs' where *: "I = x # xs" "I' = x' # xs'"
    by (fastforce simp: Suc_length_conv map_eq_Cons_conv)
  with Suc show ?case
  proof (cases m)
    case 0 thus ?thesis using Suc(3) unfolding *
      by (cases x x' rule: sum.exhaust[case_product sum.exhaust]) auto
  qed auto
qed simp

lemma stream_enc_welldef: "\<lbrakk>stream_enc (w, I) = stream_enc (w', I'); wf_formula (length I) \<phi>;
  wf_interp_for_formula (w, I) \<phi>; wf_interp_for_formula (w', I') \<phi>\<rbrakk> \<Longrightarrow>
  (w, I) \<Turnstile> \<phi> \<longleftrightarrow> (w', I') \<Turnstile> \<phi>"
proof (induction \<phi> arbitrary: w w' I I')
  case (FQ a m) thus ?case using enc_atom_welldef[of I I' m]
    by (simp split: sum.splits add: stream_map2_alt shift_snth) (metis snth_same)
next
  case (FLess m1 m2) thus ?case using enc_atom_welldef[of I I' m1] enc_atom_welldef[of I I' m2]
    by (auto split: sum.splits simp add: stream_map2_alt)
next
  case (FIn m M) thus ?case using enc_atom_welldef[of I I' m] enc_atom_welldef[of I I' M]
    by (auto split: sum.splits simp add: stream_map2_alt)
next
  case (FOr \<phi>1 \<phi>2) show ?case unfolding satisfies.simps(5)
  proof (intro disj_cong)
    from FOr(3-6) show "(w, I) \<Turnstile> \<phi>1 \<longleftrightarrow> (w', I') \<Turnstile> \<phi>1"
      by (intro FOr(1)) auto
  next
    from FOr(3-6) show "(w, I) \<Turnstile> \<phi>2 \<longleftrightarrow> (w', I') \<Turnstile> \<phi>2"
      by (intro FOr(2)) auto
  qed
next
  case (FAnd \<phi>1 \<phi>2) show ?case unfolding satisfies.simps(6)
  proof (intro conj_cong)
    from FAnd(3-6) show "(w, I) \<Turnstile> \<phi>1 \<longleftrightarrow> (w', I') \<Turnstile> \<phi>1"
      by (intro FAnd(1)) auto
  next
    from FAnd(3-6) show "(w, I) \<Turnstile> \<phi>2 \<longleftrightarrow> (w', I') \<Turnstile> \<phi>2"
      by (intro FAnd(2)) auto
  qed
next
  case (FExists \<phi>)
  hence length: "length I' = length I" by (metis length_snth_enc)
  show ?case
  proof
    assume "(w, I) \<Turnstile> FExists \<phi>"
    with FExists.prems(3) obtain p where "(w, Inl p # I) \<Turnstile> \<phi>" by auto
    moreover
    with FExists.prems have "(w', Inl p # I') \<Turnstile> \<phi>"
      apply (intro iffD1[OF FExists.IH[of w "Inl p # I" w' "Inl p # I'"]])
      apply (auto simp: stream_map2_alt split: sum.splits) []
      apply (auto split: sum.splits split_if_asm) []
      apply (blast dest!: wf_interp_for_formula_FExists)
      apply (blast dest!: wf_interp_for_formula_FExists[of I', unfolded length])
      apply assumption
      done
    ultimately show "(w', I') \<Turnstile> FExists \<phi>" by auto
  next
    assume "(w', I') \<Turnstile> FExists \<phi>"
    with FExists.prems(1,2,4) obtain p where "(w', Inl p # I') \<Turnstile> \<phi>" by auto
    moreover
    with FExists.prems have "(w, Inl p # I) \<Turnstile> \<phi>"
      apply (intro iffD2[OF FExists.IH[of w "Inl p # I" w' "Inl p # I'"]])
      apply (auto simp: stream_map2_alt split: sum.splits) []
      apply (auto split: sum.splits split_if_asm) []
      apply (blast dest!: wf_interp_for_formula_FExists)
      apply (blast dest!: wf_interp_for_formula_FExists[of I', unfolded length])
      apply assumption
      done
    ultimately show "(w, I) \<Turnstile> FExists \<phi>" by auto
  qed
next
  case (FEXISTS \<phi>)
  hence length: "length I' = length I" by (metis length_snth_enc)
  show ?case
  proof
    assume "(w, I) \<Turnstile> FEXISTS \<phi>"
    with FEXISTS.prems(3) obtain P where "finite P" "(w, Inr P # I) \<Turnstile> \<phi>" by auto
    moreover
    with FEXISTS.prems have "(w', Inr P # I') \<Turnstile> \<phi>"
      apply (intro iffD1[OF FEXISTS.IH[of w "Inr P # I" w' "Inr P # I'"]])
      apply (auto simp: stream_map2_alt split: sum.splits) []
      apply (auto split: sum.splits split_if_asm) []
      apply (blast dest!: wf_interp_for_formula_FEXISTS)
      apply (blast dest!: wf_interp_for_formula_FEXISTS[of I', unfolded length])
      apply assumption
      done
    ultimately show "(w', I') \<Turnstile> FEXISTS \<phi>" by auto
  next
    assume "(w', I') \<Turnstile> FEXISTS \<phi>"
    with FEXISTS.prems(1,2,4) obtain P where "finite P" "(w', Inr P # I') \<Turnstile> \<phi>" by auto
    moreover
    with FEXISTS.prems have "(w, Inr P # I) \<Turnstile> \<phi>"
      apply (intro iffD2[OF FEXISTS.IH[of w "Inr P # I" w' "Inr P # I'"]])
      apply (auto simp: stream_map2_alt split: sum.splits) []
      apply (auto split: sum.splits split_if_asm) []
      apply (blast dest!: wf_interp_for_formula_FEXISTS)
      apply (blast dest!: wf_interp_for_formula_FEXISTS[of I', unfolded length])
      apply assumption
      done
    ultimately show "(w, I) \<Turnstile> FEXISTS \<phi>" by auto
  qed
qed auto

lemma lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_FOr:
  assumes "wf_formula n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2)"
  shows "lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2) \<subseteq>
    (lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>1 \<union> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>2) \<inter> \<Union>{enc (w, I) | w I. length I = n \<and> wf_interp_for_formula (w, I) (FOr \<phi>\<^isub>1 \<phi>\<^isub>2)}"
    (is "_ \<subseteq> (?L1 \<union> ?L2) \<inter> ?ENC")
proof (intro equalityI subsetI)
  fix x assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2)"
  then obtain w I where
    *: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FOr \<phi>\<^isub>1 \<phi>\<^isub>2)" "length I = n" and
     "satisfies (w, I) \<phi>\<^isub>1 \<or> satisfies (w, I) \<phi>\<^isub>2" unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by auto
  thus "x \<in> (?L1 \<union> ?L2) \<inter> ?ENC"
  proof (elim disjE)
    assume "satisfies (w, I) \<phi>\<^isub>1"
    with * have "x \<in> ?L1" using assms unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by (fastforce simp del: enc.simps)
    with * show ?thesis by auto
  next
    assume "satisfies (w, I) \<phi>\<^isub>2"
    with * have "x \<in>?L2" using assms unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by (fastforce simp del: enc.simps)
    with * show ?thesis by auto
  qed
qed

lemma lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_FAnd:
  assumes "wf_formula n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2)"
  shows "lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2) \<subseteq>
    lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>1 \<inter> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>2 \<inter> \<Union>{enc (w, I) | w I. length I = n \<and> wf_interp_for_formula (w, I) (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2)}"
  using assms unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by (fastforce simp del: enc.simps)

subsection {* From WS1S to Regular expressions *}

fun rexp_of :: "nat \<Rightarrow> 'a formula \<Rightarrow> ('a \<times> bool list) rexp" where
  "rexp_of n (FQ a m) =
    Inter (TIMES [rexp.Not Zero, arbitrary_except n [(m, True)] [a], rexp.Not Zero])
    (ENC n (FQ a m))"
| "rexp_of n (FLess m1 m2) = (if m1 = m2 then Zero else
    Inter (TIMES [rexp.Not Zero, arbitrary_except n [(m1, True)] \<Sigma>,
      rexp.Not Zero, arbitrary_except n [(m2, True)] \<Sigma>,
      rexp.Not Zero]) (ENC n (FLess m1 m2)))"
| "rexp_of n (FIn m M) = 
    Inter (TIMES [rexp.Not Zero, arbitrary_except n [(min m M, True), (max m M, True)] \<Sigma>, rexp.Not Zero])
    (ENC n (FIn m M))"
| "rexp_of n (FNot \<phi>) = Inter (rexp.Not (rexp_of n \<phi>)) (ENC n (FNot \<phi>))"
| "rexp_of n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2) = Inter (Plus (rexp_of n \<phi>\<^isub>1) (rexp_of n \<phi>\<^isub>2)) (ENC n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2))"
| "rexp_of n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2) = INTERSECT [rexp_of n \<phi>\<^isub>1, rexp_of n \<phi>\<^isub>2, ENC n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2)]"
| "rexp_of n (FExists \<phi>) = samequot_exec (any, replicate n False) (Pr (rexp_of (n + 1) \<phi>))"
| "rexp_of n (FEXISTS \<phi>) = samequot_exec (any, replicate n False) (Pr (rexp_of (n + 1) \<phi>))"

fun rexp_of_alt :: "nat \<Rightarrow> 'a formula \<Rightarrow> ('a \<times> bool list) rexp" where
  "rexp_of_alt n (FQ a m) =
    TIMES [rexp.Not Zero, arbitrary_except n [(m, True)] [a], rexp.Not Zero]"
| "rexp_of_alt n (FLess m1 m2) = (if m1 = m2 then Zero else
    TIMES [rexp.Not Zero, arbitrary_except n [(m1, True)] \<Sigma>,
      rexp.Not Zero, arbitrary_except n [(m2, True)] \<Sigma>,
      rexp.Not Zero])"
| "rexp_of_alt n (FIn m M) = 
    TIMES [rexp.Not Zero, arbitrary_except n [(min m M, True), (max m M, True)] \<Sigma>, rexp.Not Zero]"
| "rexp_of_alt n (FNot \<phi>) = rexp.Not (rexp_of_alt n \<phi>)"
| "rexp_of_alt n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2) = Plus (rexp_of_alt n \<phi>\<^isub>1) (rexp_of_alt n \<phi>\<^isub>2)"
| "rexp_of_alt n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2) = Inter (rexp_of_alt n \<phi>\<^isub>1) (rexp_of_alt n \<phi>\<^isub>2)"
| "rexp_of_alt n (FExists \<phi>) = samequot_exec (any, replicate n False) (Pr (Inter (rexp_of_alt (n + 1) \<phi>) (ENC (Suc n) \<phi>)))"
| "rexp_of_alt n (FEXISTS \<phi>) = samequot_exec (any, replicate n False) (Pr (Inter (rexp_of_alt (n + 1) \<phi>) (ENC (Suc n) \<phi>)))"

definition "rexp_of' n \<phi> = Inter (rexp_of_alt n \<phi>) (ENC n \<phi>)"

lemma enc_eqI:
  assumes "x \<in> enc (w, I)" "x \<in> enc (w', I')" "wf_interp_for_formula (w, I) \<phi>" "wf_interp_for_formula (w', I') \<phi>"
    "length I = length I'"
  shows "enc (w, I) = enc (w', I')"
proof -
  from assms have "stream_enc (w, I) = stream_enc (w', I')"
    by (intro box_equals[OF _ stream_enc_enc[symmetric] stream_enc_enc[symmetric]]) auto
  thus ?thesis using assms(5) by auto
qed

lemma enc_eq_welldef:
  "\<lbrakk>enc (w, I) = enc (w', I'); wf_formula (length I) \<phi>; wf_interp_for_formula (w, I) \<phi> ;wf_interp_for_formula (w', I') \<phi>\<rbrakk> \<Longrightarrow>
  (w, I) \<Turnstile> \<phi> \<longleftrightarrow> (w', I') \<Turnstile> \<phi>"
  by (intro stream_enc_welldef) (auto simp del: stream_enc.simps intro!: enc_stream_enc)

lemma enc_welldef:
  "\<lbrakk>x \<in> enc (w, I); x \<in> enc (w', I'); length I = length I'; wf_formula (length I) \<phi>;
  wf_interp_for_formula (w, I) \<phi> ;wf_interp_for_formula (w', I') \<phi>\<rbrakk> \<Longrightarrow>
  (w, I) \<Turnstile> \<phi> \<longleftrightarrow> (w', I') \<Turnstile> \<phi>"
  by (intro enc_eq_welldef[OF enc_eqI])

lemma wf_rexp_of: "wf_formula n \<phi> \<Longrightarrow> wf n (rexp_of n \<phi>)"
  by (induct \<phi> arbitrary: n)
    (auto simp: wf_rexp_ENC intro: wf_rexp_arbitrary_except intro!: wf_samequot_exec,
     auto simp: \<sigma>_def set_n_lists image_iff)

theorem lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of: "wf_formula n \<phi> \<Longrightarrow> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi> = lang n (rexp_of n \<phi>)"
   (is "_ \<Longrightarrow> _ = ?L n \<phi>")
proof (induct \<phi> arbitrary: n)
  case (FQ a m)
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FQ a m)"
    then obtain w I where
      *: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FQ a m)" "length I = n" "(w, I) \<Turnstile> FQ a m"
      unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by blast
    hence x_alt: "x = map (split (enc_atom I)) (zip [0 ..< length x] (stake (length x) (w @- same any)))"
      by (intro encD) auto
    from FQ(1) *(2,4) obtain p where p: "I ! m = Inl p"
      by (auto simp: all_set_conv_all_nth enc_def split: sum.splits)
    with FQ(1) * have p_less: "p < length x"
      by (auto simp del: stream_enc.simps intro: trans_less_add1[OF less_length_cutSame_Inl])
    hence enc_atom: "x ! p = enc_atom I p ((w @- same any) !! p)" (is "_ = enc_atom _ _ ?p")
      by (subst x_alt, simp)
    with *(1) p_less(1) have "x = take p x @ [enc_atom I p ?p] @ drop (p + 1) x"
      using id_take_nth_drop[of p x] by auto
    moreover
    from *(2,3,4) FQ(1) p have "[enc_atom I p ?p] \<in> lang n (arbitrary_except n [(m, True)] [a])"
      by (intro enc_atom_lang_arbitrary_except_True) auto
    moreover from *(2,3) have "take p x \<in> lang n (rexp.Not Zero)"
      by (subst x_alt) (auto simp: in_set_zip shift_snth intro!: enc_atom_\<sigma> dest!: in_set_takeD)
    moreover from *(2,3) have "drop (Suc p) x \<in> lang n (rexp.Not Zero)"
      by (subst x_alt) (auto simp: in_set_zip shift_snth intro!: enc_atom_\<sigma> dest!: in_set_dropD)
    ultimately show "x \<in> ?L n (FQ a m)" using *(1,2,3)
      unfolding rexp_of.simps lang.simps(5,8) rexp_of_list.simps lang_ENC[OF FQ]
      by (auto elim: ssubst simp del: o_apply append.simps lang.simps enc.simps)
  next
    fix x let ?x = "x @- same (any, replicate n False)"
    assume x: "x \<in> ?L n (FQ a m)"
    with FQ obtain w I where
      I: "x \<in> enc (w, I)" "length I = n" "wf_interp_for_formula (w, I) (FQ a m)"
      unfolding rexp_of.simps lang.simps lang_ENC[OF FQ] by fastforce
    hence stream_enc: "stream_enc (w, I) = ?x" using stream_enc_enc by auto
    from I FQ obtain p where m: "I ! m = Inl p" "m < length I" by (auto split: sum.splits)
    with I have "wf_interp_for_formula (dec_word ?x, stream_dec n {m} ?x) (FQ a m)" unfolding I(1)
      using enc_wf_interp[OF FQ(1)[folded I(2)]] by auto
    moreover
    from x obtain u1 u u2 where "x = u1 @ u @ u2"
      "u \<in> lang n (arbitrary_except n [(m, True)] [a])"
      unfolding rexp_of.simps lang.simps rexp_of_list.simps using concE by fast
    with FQ(1) obtain v where v: "x = u1 @ [v] @ u2" "snd v ! m" "fst v = a"
      using arbitrary_except[of u n m True "[a]"] by fastforce
    from v have u: "length u1 < length x" by auto
    { from v have "snd (x ! length u1) ! m" by auto
      moreover
      from m I have "p < length x" "snd (x ! p) ! m" by (auto dest: enc_Inl simp del: enc.simps)
      moreover
      from m I have ex1: "\<exists>!p. snd (stream_enc (w, I) !! p) ! m" by (intro stream_enc_unique) auto
      ultimately have "p = length u1" unfolding stream_enc using u I(3) by auto
    } note * = this
    from v have "v = x ! length u1" by simp
    with u have "?x !! length u1 = v" by (auto simp: shift_snth)
    with * m I v(3) have "(dec_word ?x, stream_dec n {m} ?x) \<Turnstile> FQ a m"
      using stream_enc_enc[OF _ I(1), symmetric] less_length_cutSame[of w "any" "length u1" a]
      by (auto simp del: enc.simps stream_enc.simps simp: dec_word_stream_enc dest!:
        stream_dec_enc_Inl stream_dec_not_Inr split: sum.splits)
        (auto simp: stream_map2_alt cutSame_def)
    moreover from m I(2)
      have stream_enc_dec: "stream_enc (dec_word (stream_enc (w, I)), stream_dec n {m} (stream_enc (w, I))) = stream_enc (w, I)"
        by (intro stream_enc_dec)
          (auto simp: stream_map2_alt sdrop_snth shift_snth intro: stream_enc_unique,
           auto simp: stream_map2_szip stream.set_natural')
    moreover from I have "wf_word n x" unfolding wf_word by (auto elim: enc_set_\<sigma> simp del: enc.simps)
    ultimately show "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FQ a m)" unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def using m I(1,3)
        by (auto simp del: enc.simps stream_enc.simps intro!: exI[of _ "enc (dec_word ?x, stream_dec n {m} ?x)"],
          fastforce simp del: enc.simps stream_enc.simps,
          auto simp del: stream_enc.simps simp: stream_enc[symmetric] I(2))
  qed
next
  case (FLess m m')
  show ?case
  proof (cases "m = m'")
    case False
    thus ?thesis
    proof (intro equalityI subsetI)
      fix x assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FLess m m')"
      then obtain w I where
        *: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FLess m m')" "length I = n" "(w, I) \<Turnstile> FLess m m'"
        unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by blast
      hence x_alt: "x = map (split (enc_atom I)) (zip [0 ..< length x] (stake (length x) (w @- same any)))"
        by (intro encD) auto
      from FLess(1) *(2,4) obtain p q where pq: "I ! m = Inl p" "I ! m' = Inl q" "p < q"
        by (auto simp: all_set_conv_all_nth enc_def split: sum.splits)
      with FLess(1) *(1,2,3) have pq_less: "p < length x" "q < length x"
        by (auto simp del: stream_enc.simps intro!: trans_less_add1[OF less_length_cutSame_Inl])
      hence enc_atom: "x ! p = enc_atom I p ((w @- same any) !! p)" (is "_ = enc_atom _ _ ?p")
                      "x ! q = enc_atom I q ((w @- same any) !! q)" (is "_ = enc_atom _ _ ?q") by (subst x_alt, simp)+
      with *(1) pq_less(1) have "x = take p x @ [enc_atom I p ?p] @ drop (p + 1) x"
        using id_take_nth_drop[of p x] by auto
      also have "drop (p + 1) x = take (q - p - 1) (drop (p + 1) x) @
        [enc_atom I q ?q] @ drop (q - p) (drop (p + 1) x)" (is "_ = ?LHS")
        using id_take_nth_drop[of "q - p - 1" "drop (p + 1) x"] pq pq_less(2) enc_atom(2) by auto
      finally have "x = take p x @ [enc_atom I p ?p] @ ?LHS" .
      moreover from *(2,3) FLess(1) pq(1)
       have "[enc_atom I p ?p] \<in> lang n (arbitrary_except n [(m, True)] \<Sigma>)"
         by (intro enc_atom_lang_arbitrary_except_True) (auto simp: shift_snth)
      moreover from *(2,3) FLess(1) pq(2)
       have "[enc_atom I q ?q] \<in> lang n (arbitrary_except n [(m', True)] \<Sigma>)"
         by (intro enc_atom_lang_arbitrary_except_True) (auto simp: shift_snth)
      moreover from *(2,3) have "take p x \<in> lang n (rexp.Not Zero)"
        by (subst x_alt) (auto simp: in_set_zip shift_snth intro!: enc_atom_\<sigma> dest!: in_set_takeD)
      moreover from *(2,3) have "take (q - p - 1) (drop (Suc p) x) \<in> lang n (rexp.Not Zero)"
        by (subst x_alt) (auto simp: in_set_zip shift_snth intro!: enc_atom_\<sigma> dest!: in_set_dropD in_set_takeD)
      moreover from *(2,3) have "drop (q - p) (drop (Suc p) x) \<in> lang n (rexp.Not Zero)"
        by (subst x_alt) (auto simp: in_set_zip shift_snth intro!: enc_atom_\<sigma> dest!: in_set_dropD)
      ultimately show "x \<in> ?L n (FLess m m')" using *(1,2,3)
        unfolding rexp_of.simps lang.simps(5,8) rexp_of_list.simps Int_Diff lang_ENC[OF FLess] if_not_P[OF False]
        by (auto elim: ssubst simp del: o_apply append.simps lang.simps enc.simps)
    next
      fix x let ?x = "x @- same (any, replicate n False)"
      assume x: "x \<in> ?L n (FLess m m')"
      with FLess obtain w I where
        I: "x \<in> enc (w, I)" "length I = n" "wf_interp_for_formula (w, I) (FLess m m')"
        unfolding rexp_of.simps lang.simps lang_ENC[OF FLess] if_not_P[OF False] by fastforce
      hence stream_enc: "stream_enc (w, I) = x @- same (any, replicate n False)" using stream_enc_enc by auto
      from I FLess obtain p p' where m: "I ! m = Inl p" "m < length I" "I ! m' = Inl p'" "m' < length I"
        by (auto split: sum.splits)
      with I have "wf_interp_for_formula (dec_word ?x, stream_dec n {m, m'} ?x) (FLess m m')" unfolding I(1)
        using enc_wf_interp[OF FLess(1)[folded I(2)]] by auto
      moreover
      from x obtain u1 u u2 u' u3 where "x = u1 @ u @ u2 @ u' @ u3"
        "u \<in> lang n (arbitrary_except n [(m, True)] \<Sigma>)" 
        "u' \<in> lang n (arbitrary_except n [(m', True)] \<Sigma>)"
        unfolding rexp_of.simps lang.simps rexp_of_list.simps if_not_P[OF False] using concE by fast
      with FLess(1) obtain v v' where v: "x = u1 @ [v] @ u2 @ [v'] @ u3"
         "snd v ! m" "snd v' ! m'" "fst v \<in> set \<Sigma>" "fst v' \<in> set \<Sigma>"
        using arbitrary_except[of u n m True \<Sigma>] arbitrary_except[of u' n m' True \<Sigma>] by fastforce
      hence u: "length u1 < length x" and u': "Suc (length u1 + length u2) < length x" (is "?u' < _") by auto
      { from v have "snd (x ! length u1) ! m" by auto
        moreover
        from m I have "p < length x" "snd (x ! p) ! m" by (auto dest: enc_Inl simp del: enc.simps)
        moreover
        from m I have ex1: "\<exists>!p. snd (stream_enc (w, I) !! p) ! m" by (intro stream_enc_unique) auto
        ultimately have "p = length u1" unfolding stream_enc using u I(3) by auto
      }
      { from v have "snd (x ! ?u') ! m'" by (auto simp: nth_append)
        moreover
        from m I have "p' < length x" "snd (x ! p') ! m'" by (auto dest: enc_Inl simp del: enc.simps)
        moreover
        from m I have ex1: "\<exists>!p. snd (stream_enc (w, I) !! p) ! m'" unfolding I(1) by (intro stream_enc_unique) auto
        ultimately have "p' = ?u'" unfolding stream_enc using u' I(3) by auto (metis shift_snth_less)
      } note * = this `p = length u1`
      with m I have "(dec_word ?x, stream_dec n {m, m'} ?x) \<Turnstile> FLess m m'"
        using stream_enc_enc[OF _ I(1), symmetric]
        by (auto dest: stream_dec_not_Inr stream_dec_enc_Inl split: sum.splits simp del: stream_enc.simps)
      moreover from m I(2)
        have stream_enc_dec: "stream_enc (dec_word (stream_enc (w, I)), stream_dec n {m, m'} (stream_enc (w, I))) = stream_enc (w, I)"
        by (intro stream_enc_dec)
          (auto simp: stream_map2_alt sdrop_snth shift_snth intro: stream_enc_unique,
           auto simp: stream_map2_szip stream.set_natural')
      moreover from I have "wf_word n x" unfolding wf_word by (auto elim: enc_set_\<sigma> simp del: enc.simps)
      ultimately show "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FLess m m')" unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def using m I(1,3)
        by (auto simp del: enc.simps stream_enc.simps intro!: exI[of _ "enc (dec_word ?x, stream_dec n {m, m'} ?x)"],
          fastforce simp del: enc.simps stream_enc.simps,
          auto simp del: stream_enc.simps simp: stream_enc[symmetric] I(2))
    qed
  qed (simp add: lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def del: o_apply)
next
  case (FIn m M)
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FIn m M)"
    then obtain w I where
      *: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FIn m M)" "length I = n" "(w, I) \<Turnstile> FIn m M"
      unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by blast
    hence x_alt: "x = map (split (enc_atom I)) (zip [0 ..< length x] (stake (length x) (w @- same any)))"
      by (intro encD) auto
    from FIn(1) *(2,4) obtain p P where p: "I ! m = Inl p" "I ! M = Inr P" "p \<in> P"
      by (auto simp: all_set_conv_all_nth enc_def split: sum.splits)
    with FIn(1) *(1,2,3) have p_less: "p < length x" "\<forall>p \<in> P. p < length x"
      by (auto simp del: stream_enc.simps intro: trans_less_add1[OF less_length_cutSame_Inl]
        trans_less_add1[OF bspec[OF less_length_cutSame_Inr]])
    hence enc_atom: "x ! p = enc_atom I p ((w @- same any) !! p)" (is "_ = enc_atom _ _ ?p")
            "\<forall>p \<in> P. x ! p = enc_atom I p ((w @- same any) !! p)" (is "Ball _ (\<lambda>p. _ = enc_atom _ _ (?P p))")
      by (subst x_alt, simp)+
    with *(1) p_less(1) have "x = take p x @ [enc_atom I p ?p] @ drop (p + 1) x"
      using id_take_nth_drop[of p x] by auto
    moreover
    have "[enc_atom I p ?p] \<in> lang n (arbitrary_except n [(min m M, True), (max m M, True)] \<Sigma>)"
    proof (cases "m < M")
      case True with *(2,3) FIn(1) p show ?thesis
        by (intro enc_atom_lang_arbitrary_except_True2) (auto simp: min_absorb1 max_absorb2 shift_snth)
    next
      case False with *(2,3) FIn(1) p show ?thesis
        by (intro enc_atom_lang_arbitrary_except_True2) (auto simp: min_absorb2 max_absorb1 shift_snth)
    qed
    moreover from *(2,3) have "take p x \<in> lang n (rexp.Not Zero)"
      by (subst x_alt) (auto simp: in_set_zip shift_snth intro!: enc_atom_\<sigma> dest!: in_set_takeD)
    moreover from *(2,3) have "drop (Suc p) x \<in> lang n (rexp.Not Zero)"
      by (subst x_alt) (auto simp: in_set_zip shift_snth intro!: enc_atom_\<sigma> dest!: in_set_dropD)
    ultimately show "x \<in> ?L n (FIn m M)" using *(1,2,3)
      unfolding rexp_of.simps lang.simps(5,8) rexp_of_list.simps Int_Diff lang_ENC[OF FIn]
      by (auto elim: ssubst simp del: o_apply append.simps lang.simps enc.simps)
  next
    fix x let ?x = "x @- same (any, replicate n False)"
    assume x: "x \<in> ?L n (FIn m M)"
    with FIn obtain w I where
      I: "x \<in> enc (w, I)" "length I = n" "wf_interp_for_formula (w, I) (FIn m M)"
      unfolding rexp_of.simps lang.simps lang_ENC[OF FIn] by fastforce
    hence stream_enc: "stream_enc (w, I) = ?x" using stream_enc_enc by auto
    from I FIn obtain p P where m: "I ! m = Inl p" "m < length I" "I ! M = Inr P" "M < length I"
      by (auto split: sum.splits)
    with I have "wf_interp_for_formula (dec_word ?x, stream_dec n {m} ?x) (FIn m M)" unfolding I(1)
      using enc_wf_interp[OF FIn(1)[folded I(2)]] by auto
    moreover
    from x obtain u1 u u2 where "x = u1 @ u @ u2"
      "u \<in> lang n (arbitrary_except n [(min m M, True), (max m M, True)] \<Sigma>)"
      unfolding rexp_of.simps lang.simps rexp_of_list.simps using concE by fast
    with FIn(1) obtain v where v: "x = u1 @ [v] @ u2" and "snd v ! min m M" "snd v ! max m M" "fst v \<in> set \<Sigma>"
      using arbitrary_except2[of u n "min m M" True "max m M" True \<Sigma>] by fastforce
    hence v': "snd v ! m" "snd v ! M"
      by (induct "m < M") (auto simp: min_absorb1 min_absorb2 max_absorb1 max_absorb2)
    from v have u: "length u1 < length x" by auto
    { from v v' have "snd (x ! length u1) ! m" by auto
      moreover
      from m I have "p < length x" "snd (x ! p) ! m" by (auto dest: enc_Inl simp del: enc.simps)
      moreover
      from m I have ex1: "\<exists>!p. snd (stream_enc (w, I) !! p) ! m" by (intro stream_enc_unique) auto
      ultimately have "p = length u1" unfolding stream_enc using u I(3) by auto
    } note * = this
    from v v' have "v = x ! length u1" by simp
    with v'(2) m(3,4) u I(1,3) have "length u1 \<in> P" by (auto dest!: enc_Inr simp del: enc.simps)
    with * m I have "(dec_word ?x, stream_dec n {m} ?x) \<Turnstile> FIn m M"
      using stream_enc_enc[OF _ I(1), symmetric]
      by (auto simp del: stream_enc.simps dest: stream_dec_not_Inr stream_dec_not_Inl
        stream_dec_enc_Inl stream_dec_enc_Inr split: sum.splits)
    moreover from m I(2)
      have stream_enc_dec: "stream_enc (dec_word (stream_enc (w, I)), stream_dec n {m} (stream_enc (w, I))) = stream_enc (w, I)"
        by (intro stream_enc_dec)
          (auto simp: stream_map2_alt sdrop_snth shift_snth intro: stream_enc_unique,
           auto simp: stream_map2_szip stream.set_natural')
    moreover from I have "wf_word n x" unfolding wf_word by (auto elim: enc_set_\<sigma> simp del: enc.simps)
    ultimately show "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FIn m M)" unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def using m I(1,3)
        by (auto simp del: enc.simps stream_enc.simps intro!: exI[of _ "enc (dec_word ?x, stream_dec n {m} ?x)"],
          fastforce simp del: enc.simps stream_enc.simps,
          auto simp del: stream_enc.simps simp: stream_enc[symmetric] I(2))
  qed
next
  case (FOr \<phi>\<^isub>1 \<phi>\<^isub>2)
  from FOr(3) have IH1: "lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>1 = lang n (rexp_of n \<phi>\<^isub>1)"
    by (intro FOr(1)) auto
  from FOr(3) have IH2: "lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>2 = lang n (rexp_of n \<phi>\<^isub>2)"
    by (intro FOr(2)) auto
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2)" thus "x \<in> lang n (rexp_of n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2))"
      using lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_FOr[OF FOr(3)] unfolding lang_ENC[OF FOr(3)] rexp_of.simps lang.simps IH1 IH2 by blast
  next
    fix x assume "x \<in> lang n (rexp_of n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2))"
    then obtain w I where or: "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>1 \<or> x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>2" and I: "x \<in> enc (w, I)" "length I = n"
      "wf_interp_for_formula (w, I) (FOr \<phi>\<^isub>1 \<phi>\<^isub>2)"
      unfolding lang_ENC[OF FOr(3)] rexp_of.simps lang.simps IH1 IH2 Int_Diff by auto
    have "(w, I) \<Turnstile> \<phi>\<^isub>1 \<or> (w, I) \<Turnstile> \<phi>\<^isub>2"
    proof (intro mp[OF disj_mono[OF impI impI] or])
      assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>1"
      with I FOr(3) show "(w, I) \<Turnstile> \<phi>\<^isub>1"
        unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def I(1) wf_interp_for_formula_FOr
        by (auto dest!: enc_welldef[of x w I _ _ \<phi>\<^isub>1] simp del: enc.simps)
    next
      assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>2"
      with I FOr(3) show "(w, I) \<Turnstile> \<phi>\<^isub>2"
        unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def I(1) wf_interp_for_formula_FOr
        by (auto dest!: enc_welldef[of x w I _ _ \<phi>\<^isub>2] simp del: enc.simps)
    qed
    with I show "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FOr \<phi>\<^isub>1 \<phi>\<^isub>2)" unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by auto
  qed
next
  case (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2)
  from FAnd(3) have IH1: "lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>1 = lang n (rexp_of n \<phi>\<^isub>1)"
    by (intro FAnd(1)) auto
  from FAnd(3) have IH2: "lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>2 = lang n (rexp_of n \<phi>\<^isub>2)"
    by (intro FAnd(2)) auto
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2)" thus "x \<in> lang n (rexp_of n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2))"
      using lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_FAnd[OF FAnd(3)]
      unfolding lang_ENC[OF FAnd(3)] rexp_of.simps rexp_of_list.simps lang.simps IH1 IH2 Int_assoc
      by blast
  next
    fix x assume "x \<in> lang n (rexp_of n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2))"
    then obtain w I where "and": "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>1 \<and> x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>2" and I: "x \<in> enc (w, I)" "length I = n"
      "wf_interp_for_formula (w, I) (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2)"
      unfolding lang_ENC[OF FAnd(3)] rexp_of.simps rexp_of_list.simps lang.simps IH1 IH2 Int_Diff by auto
    have "(w, I) \<Turnstile> \<phi>\<^isub>1 \<and> (w, I) \<Turnstile> \<phi>\<^isub>2"
    proof (intro mp[OF conj_mono[OF impI impI] "and"])
      assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>1"
      with I FAnd(3) show "(w, I) \<Turnstile> \<phi>\<^isub>1"
        unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def I(1) wf_interp_for_formula_FAnd
        by (auto dest!: enc_welldef[of x w I _ _ \<phi>\<^isub>1] simp del: enc.simps)
    next
      assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>\<^isub>2"
      with I FAnd(3) show "(w, I) \<Turnstile> \<phi>\<^isub>2"
        unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def I(1) wf_interp_for_formula_FAnd
        by (auto dest!: enc_welldef[of x w I _ _ \<phi>\<^isub>2] simp del: enc.simps)
    qed
    with I show "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2)" unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by auto
  qed
next
  case (FNot \<phi>)
  hence IH: "?L n \<phi> =  lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>" by simp
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FNot \<phi>)"
    then obtain w I where
      *: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) \<phi>" "length I = n" and unsat: "\<not> (w, I) \<Turnstile> \<phi>"
       unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by auto
    { assume "x \<in> ?L n \<phi>"
      hence "(w, I) \<Turnstile> \<phi>" using enc_welldef[of x w I _ _ \<phi>, OF *(1) _ _ _ *(2)] FNot(2)
        unfolding *(3) lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def IH by auto
    }
    with unsat have "x \<notin> ?L n \<phi>" by blast
    with * show "x \<in> ?L n (FNot \<phi>)" unfolding rexp_of.simps lang.simps using lang_ENC[OF FNot(2)]
      by (auto simp del: enc.simps simp: comp_def intro: enc_set_\<sigma>)
  next
    fix x assume "x \<in> ?L n (FNot \<phi>)"
    with IH have "x \<in> lang n (ENC n (FNot \<phi>))" and x: "x \<notin> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>" by (auto simp del: o_apply)
    then obtain w I where *: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FNot \<phi>)" "length I = n"
      unfolding lang_ENC[OF FNot(2)] by blast
    { assume "\<not> (w, I) \<Turnstile> FNot \<phi>"
      with * have "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi>" unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by auto
    }
    with x * show "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FNot \<phi>)"  unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by blast
  qed
next
  case (FExists \<phi>)
  have \<sigma>: "(any, replicate n False) \<in> (set o \<sigma> \<Sigma>) n" by (auto simp: \<sigma>_def set_n_lists image_iff)
  from FExists(2) have wf: "wf n (Pr (rexp_of (Suc n) \<phi>))" by (fastforce intro: wf_rexp_of)
  note lang_quot = lang_samequot_exec[OF wf \<sigma>]
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FExists \<phi>)"
    then obtain w I p where
      *: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FExists \<phi>)" "length I = n" "(w, Inl p # I) \<Turnstile> \<phi>"
       unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by auto
    with FExists(2) have "enc (w, Inl p # I) \<subseteq> ?L (Suc n) \<phi>"
      by (subst FExists(1)[of "Suc n", symmetric])
        (fastforce simp del: enc.simps simp: lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def nth_Cons' intro!: exI[of _ "enc (w, Inl p # I)"])+
    thus "x \<in> ?L n (FExists \<phi>)" using *(1,2,3)
      by (auto simp: lang_quot simp del: o_apply enc.simps elim: set_mp[OF SAMEQUOT_mono[OF image_mono]])
  next
    fix x assume "x \<in> ?L n (FExists \<phi>)"
    then obtain x' m where "x' \<in> ?L (Suc n) \<phi>" and
      x: "x = fin_cutSame (any, replicate n False) (map \<pi> x') @ replicate m (any, replicate n False)" 
      by (auto simp: lang_quot SAMEQUOT_def simp del: o_apply enc.simps)
    with FExists(2) have "x' \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S (Suc n) \<phi>"
      by (intro subsetD[OF equalityD2[OF FExists(1)], of "Suc n" x'])
        (auto split: split_if_asm sum.splits)
    then obtain w I' where
      *: "x' \<in> enc (w, I')" "wf_interp_for_formula (w, I') \<phi>" "length I' = Suc n" "(w, I') \<Turnstile> \<phi>"
       unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by blast
    moreover then obtain I\<^isub>0 I where "I' = I\<^isub>0 # I" by (cases I') auto
    moreover with FExists(2) *(2) obtain p where "I\<^isub>0 = Inl p"
      by (auto simp: nth_Cons' split: sum.splits split_if_asm)
    ultimately have "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FExists \<phi>)" "length I = n"
      "(w, I) \<Turnstile> FExists \<phi>" using FExists(2) fin_cutSame_tl[OF ex_Loop_stream_enc, of "Inl p # I" w]
      unfolding x by (auto simp add: \<pi>_def nth_Cons' split: split_if_asm)
    thus "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FExists \<phi>)" unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by (auto intro!: exI[of _ I])
  qed
next
  case (FEXISTS \<phi>)
  have \<sigma>: "(any, replicate n False) \<in> (set o \<sigma> \<Sigma>) n" by (auto simp: \<sigma>_def set_n_lists image_iff)
  from FEXISTS(2) have wf: "wf n (Pr (rexp_of (Suc n) \<phi>))" by (fastforce intro: wf_rexp_of)
  note lang_quot = lang_samequot_exec[OF wf \<sigma>]
  show ?case
  proof (intro equalityI subsetI)
    fix x assume "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FEXISTS \<phi>)"
    then obtain w I P where
      *: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FEXISTS \<phi>)" "length I = n" "finite P" "(w, Inr P # I) \<Turnstile> \<phi>"
       unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by auto
    with FEXISTS(2) have "enc (w, Inr P # I) \<subseteq> ?L (Suc n) \<phi>"
      by (subst FEXISTS(1)[of "Suc n", symmetric])
        (fastforce simp del: enc.simps simp: lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def nth_Cons' intro!: exI[of _ "enc (w, Inr P # I)"])+
    thus "x \<in> ?L n (FEXISTS \<phi>)" using *(1,2,3,4)
      by (auto simp: lang_quot simp del: o_apply enc.simps elim: set_mp[OF SAMEQUOT_mono[OF image_mono]])
  next
    fix x assume "x \<in> ?L n (FEXISTS \<phi>)"
    then obtain x' m where "x' \<in> ?L (Suc n) \<phi>" and
      x: "x = fin_cutSame (any, replicate n False) (map \<pi> x') @ replicate m (any, replicate n False)" 
      by (auto simp: lang_quot SAMEQUOT_def simp del: o_apply enc.simps)
    with FEXISTS(2) have "x' \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S (Suc n) \<phi>"
      by (intro subsetD[OF equalityD2[OF FEXISTS(1)], of "Suc n" x'])
        (auto split: split_if_asm sum.splits)
    then obtain w I' where
      *: "x' \<in> enc (w, I')" "wf_interp_for_formula (w, I') \<phi>" "length I' = Suc n" "(w, I') \<Turnstile> \<phi>"
       unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by blast
    moreover then obtain I\<^isub>0 I where "I' = I\<^isub>0 # I" by (cases I') auto
    moreover with FEXISTS(2) *(2) obtain P where "I\<^isub>0 = Inr P" "finite P"
      by (auto simp: nth_Cons' split: sum.splits split_if_asm)
    ultimately have "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FEXISTS \<phi>)" "length I = n"
      "(w, I) \<Turnstile> FEXISTS \<phi>"  using FEXISTS(2) fin_cutSame_tl[OF ex_Loop_stream_enc, of "Inr P # I"]
      unfolding x by (auto simp: nth_Cons' \<pi>_def split: split_if_asm)
    thus "x \<in> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n (FEXISTS \<phi>)" unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_def by (auto intro!: exI[of _ I])
  qed
qed

lemma wf_rexp_of_alt: "wf_formula n \<phi> \<Longrightarrow> wf n (rexp_of_alt n \<phi>)"
  by (induct \<phi> arbitrary: n)
    (auto simp: wf_rexp_ENC intro: wf_rexp_arbitrary_except intro!: wf_samequot_exec,
     auto simp: \<sigma>_def set_n_lists image_iff)

lemma wf_rexp_of': "wf_formula n \<phi> \<Longrightarrow> wf n (rexp_of' n \<phi>)"
  unfolding rexp_of'_def by (auto intro: wf_rexp_of_alt wf_rexp_ENC)

lemma ENC_FNot: "ENC n (FNot \<phi>) = ENC n \<phi>"
  unfolding ENC_def by auto

lemma ENC_FAnd:
  "wf_formula n (FAnd \<phi> \<psi>) \<Longrightarrow> lang n (ENC n (FAnd \<phi> \<psi>)) \<subseteq> lang n (ENC n \<phi>) \<inter> lang n (ENC n \<psi>)"
proof
  fix x assume wf: "wf_formula n (FAnd \<phi> \<psi>)" and x: "x \<in> lang n (ENC n (FAnd \<phi> \<psi>))"
  hence wf1: "wf_formula n \<phi>" and wf2: "wf_formula n \<psi>" by auto
  from x obtain w I where I: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FAnd \<phi> \<psi>)" "length I = n"
    using lang_ENC[OF wf] by auto
  hence "wf_interp_for_formula (w, I) \<phi>" "wf_interp_for_formula (w, I) \<psi>"
    using wf_interp_for_formula_FAnd by auto
  thus "x \<in> lang n (ENC n \<phi>) \<inter> lang n (ENC n \<psi>)"
    unfolding lang_ENC[OF wf1] lang_ENC[OF wf2] using I by blast
qed
  
lemma ENC_FOr:
  "wf_formula n (FOr \<phi> \<psi>) \<Longrightarrow> lang n (ENC n (FOr \<phi> \<psi>)) \<subseteq> lang n (ENC n \<phi>) \<inter> lang n (ENC n \<psi>)"
proof
  fix x assume wf: "wf_formula n (FOr \<phi> \<psi>)" and x: "x \<in> lang n (ENC n (FOr \<phi> \<psi>))"
  hence wf1: "wf_formula n \<phi>" and wf2: "wf_formula n \<psi>" by auto
  from x obtain w I where I: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FOr \<phi> \<psi>)" "length I = n"
    using lang_ENC[OF wf] by auto
  hence "wf_interp_for_formula (w, I) \<phi>" "wf_interp_for_formula (w, I) \<psi>"
    using wf_interp_for_formula_FOr by auto
  thus "x \<in> lang n (ENC n \<phi>) \<inter> lang n (ENC n \<psi>)"
    unfolding lang_ENC[OF wf1] lang_ENC[OF wf2] using I by blast
qed

lemma ENC_FExists:
  "wf_formula n (FExists \<phi>) \<Longrightarrow> lang n (ENC n (FExists \<phi>)) =
  SAMEQUOT (any, replicate n False) (map \<pi> ` lang (Suc n) (ENC (Suc n) \<phi>))" (is "_ \<Longrightarrow> ?L = ?R")
proof (intro equalityI subsetI)
  fix x assume wf: "wf_formula n (FExists \<phi>)" and x: "x \<in> ?L"
  hence wf1: "wf_formula (Suc n) \<phi>" by auto
  from x obtain w I where I: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FExists \<phi>)" "length I = n"
    using lang_ENC[OF wf] by auto
  from I(2) obtain p where "wf_interp_for_formula (w, Inl p # I) \<phi>"
    using wf_interp_for_formula_FExists[OF wf[folded I(3)]] by blast
  with I(3) show "x \<in> ?R"
    unfolding lang_ENC[OF wf1] using I(1) tl_enc[of "Inl p" I, symmetric]
    by (simp del: enc.simps) 
      (fastforce simp del: enc.simps elim!: set_rev_mp[OF _ SAMEQUOT_mono[OF image_mono]]
        intro: exI[of _ "enc (w, Inl p # I)"])
next
  fix x assume wf: "wf_formula n (FExists \<phi>)" and x: "x \<in> ?R"
  hence wf1: "wf_formula (Suc n) \<phi>" and "0 \<in> FOV \<phi>" by auto
  from x obtain w I where I: "x \<in> SAMEQUOT (any, replicate n False) (map \<pi> ` enc (w, I))"
    "wf_interp_for_formula (w, I) \<phi>" "length I = Suc n"
    using lang_ENC[OF wf1] unfolding SAMEQUOT_def by fast
  with `0 \<in> FOV \<phi>` obtain p I' where I': "I = Inl p # I'" by (cases I) (fastforce split: sum.splits)+
  with I have wtlI: "x \<in> enc (w, I')" "length I' = n" using tl_enc[of "Inl p" I' w] by auto
  have "wf_interp_for_formula (w, I') (FExists \<phi>)"
    using wf_interp_for_formula_FExists[OF wf[folded wtlI(2)]]
          wf_interp_for_formula_any_Inl[OF I(2)[unfolded I']] ..
  with wtlI show "x \<in> ?L" unfolding lang_ENC[OF wf] by blast
qed

lemma ENC_FEXISTS:
  "wf_formula n (FEXISTS \<phi>) \<Longrightarrow> lang n (ENC n (FEXISTS \<phi>)) =
  SAMEQUOT (any, replicate n False) (map \<pi> ` lang (Suc n) (ENC (Suc n) \<phi>))" (is "_ \<Longrightarrow> ?L = ?R")
proof (intro equalityI subsetI)
  fix x assume wf: "wf_formula n (FEXISTS \<phi>)" and x: "x \<in> ?L"
  hence wf1: "wf_formula (Suc n) \<phi>" by auto
  from x obtain w I where I: "x \<in> enc (w, I)" "wf_interp_for_formula (w, I) (FEXISTS \<phi>)" "length I = n"
    using lang_ENC[OF wf] by auto
  from I(2) obtain P where "wf_interp_for_formula (w, Inr P # I) \<phi>"
    using wf_interp_for_formula_FEXISTS[OF wf[folded I(3)]] by blast
  with I(3) show "x \<in> ?R"
    unfolding lang_ENC[OF wf1] using I(1) tl_enc[of "Inr P" I, symmetric]
    by (simp del: enc.simps) 
      (fastforce simp del: enc.simps elim!: set_rev_mp[OF _ SAMEQUOT_mono[OF image_mono]]
        intro: exI[of _ "enc (w, Inr P # I)"])
next
  fix x assume wf: "wf_formula n (FEXISTS \<phi>)" and x: "x \<in> ?R"
  hence wf1: "wf_formula (Suc n) \<phi>" and "0 \<in> SOV \<phi>" by auto
  from x obtain w I where I: "x \<in> SAMEQUOT (any, replicate n False) (map \<pi> ` enc (w, I))"
    "wf_interp_for_formula (w, I) \<phi>" "length I = Suc n"
    using lang_ENC[OF wf1] unfolding SAMEQUOT_def by fast
  with `0 \<in> SOV \<phi>` obtain P I' where I': "I = Inr P # I'" by (cases I) (fastforce split: sum.splits)+
  with I have wtlI: "x \<in> enc (w, I')" "length I' = n" using tl_enc[of "Inr P" I' w] by auto
  have "wf_interp_for_formula (w, I') (FEXISTS \<phi>)"
    using wf_interp_for_formula_FEXISTS[OF wf[folded wtlI(2)]]
          wf_interp_for_formula_any_Inr[OF I(2)[unfolded I']] ..
  with wtlI show "x \<in> ?L" unfolding lang_ENC[OF wf] by blast
qed

lemma lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of_rexp_of':
  "wf_formula n \<phi> \<Longrightarrow> lang n (rexp_of n \<phi>) = lang n (rexp_of' n \<phi>)"
unfolding rexp_of'_def proof (induction \<phi> arbitrary: n)
  case (FNot \<phi>)
  hence "wf_formula n \<phi>" by simp
  with FNot.IH show ?case unfolding rexp_of.simps rexp_of_alt.simps lang.simps ENC_FNot by blast
next
  case (FAnd \<phi>\<^isub>1 \<phi>\<^isub>2)
  hence wf1: "wf_formula n \<phi>\<^isub>1" and wf2: "wf_formula n \<phi>\<^isub>2" by force+
  from FAnd.IH(1)[OF wf1] FAnd.IH(2)[OF wf2] show ?case using ENC_FAnd[OF FAnd.prems]
    unfolding rexp_of.simps rexp_of_alt.simps lang.simps rexp_of_list.simps by blast
next
  case (FOr \<phi>\<^isub>1 \<phi>\<^isub>2)
  hence wf1: "wf_formula n \<phi>\<^isub>1" and wf2: "wf_formula n \<phi>\<^isub>2" by force+
  from FOr.IH(1)[OF wf1] FOr.IH(2)[OF wf2] show ?case using ENC_FOr[OF FOr.prems]
    unfolding rexp_of.simps rexp_of_alt.simps lang.simps by blast
next
  case (FExists \<phi>)
  from FExists(2) have IH: "lang (n + 1) (rexp_of (n + 1) \<phi>) =
    lang (n + 1) (Inter (rexp_of_alt (n + 1) \<phi>) (ENC (n + 1) \<phi>))" by (intro FExists.IH) auto
  have \<sigma>: "(any, replicate n False) \<in> (set o \<sigma> \<Sigma>) n" by (auto simp: \<sigma>_def set_n_lists image_iff)
  from FExists(2) have wf: "wf n (Pr (rexp.Inter (rexp_of_alt (n + 1) \<phi>) (ENC (n + 1) \<phi>)))"
    "wf n (Pr (rexp_of (n + 1) \<phi>))" by (fastforce intro!: wf_rexp_of wf_rexp_of_alt wf_rexp_ENC)+
  note lang_quot = lang_samequot_exec[OF wf(1) \<sigma>] lang_samequot_exec[OF wf(2) \<sigma>]
  show ?case unfolding rexp_of.simps rexp_of_alt.simps lang.simps IH lang_quot Suc_eq_plus1
    ENC_FExists[OF FExists.prems, unfolded Suc_eq_plus1] by (auto simp add: SAMEQUOT_def)
next
  case (FEXISTS \<phi>)
  from FEXISTS(2) have IH: "lang (n + 1) (rexp_of (n + 1) \<phi>) =
    lang (n + 1) (Inter (rexp_of_alt (n + 1) \<phi>) (ENC (n + 1) \<phi>))" by (intro FEXISTS.IH) auto
  have \<sigma>: "(any, replicate n False) \<in> (set o \<sigma> \<Sigma>) n" by (auto simp: \<sigma>_def set_n_lists image_iff)
  from FEXISTS(2) have wf: "wf n (Pr (rexp.Inter (rexp_of_alt (n + 1) \<phi>) (ENC (n + 1) \<phi>)))"
    "wf n (Pr (rexp_of (n + 1) \<phi>))" by (fastforce intro: wf_rexp_of wf_rexp_of_alt wf_rexp_ENC)+
  note lang_quot = lang_samequot_exec[OF wf(1) \<sigma>] lang_samequot_exec[OF wf(2) \<sigma>]
  show ?case unfolding rexp_of.simps rexp_of_alt.simps lang.simps IH lang_quot Suc_eq_plus1
    ENC_FEXISTS[OF FEXISTS.prems, unfolded Suc_eq_plus1] by (auto simp add: SAMEQUOT_def)
qed auto

theorem lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of': "wf_formula n \<phi> \<Longrightarrow> lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S n \<phi> = lang n (rexp_of' n \<phi>)"
  unfolding lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of_rexp_of'[symmetric] by (rule lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of)

end

end
