(* Author: Dmitriy Traytel *)

header {* Deciding Equivalence of WS1S Formulas *}

(*<*)
theory WS1S_Equivalence_Checking
imports Smart_Constructors_Normalization WS1S_Normalization "~~/src/HOL/Library/Option_ord"
begin
(*>*)

type_synonym 'a T = "'a \<times> bool list"
abbreviation "\<LL> \<equiv> \<lambda>\<Sigma>. project.lang (set o (\<sigma> \<Sigma>)) \<pi>"

definition wf_rexp where [code del]:
  "wf_rexp \<Sigma> = alphabet.wf (set o \<sigma> \<Sigma>)"

interpretation project "set o \<sigma> \<Sigma>" \<pi>
  where "alphabet.wf (set o \<sigma> \<Sigma>) = wf_rexp \<Sigma>"
  by (unfold_locales) (auto simp: \<sigma>_def \<pi>_def wf_rexp_def)

definition norm_lderiv where [code del]: "norm_lderiv \<equiv> \<lambda>\<Sigma>. embed.lderiv (\<epsilon> \<Sigma>)"
definition norm_rderiv where [code del]: "norm_rderiv \<equiv> \<lambda>\<Sigma>. embed.rderiv (\<epsilon> \<Sigma>)"
definition norm_rderiv_and_add where [code del]: "norm_rderiv_and_add \<equiv> \<lambda>\<Sigma>. embed.rderiv_and_add (\<epsilon> \<Sigma>)"
definition quot where [code del]: "quot \<equiv> \<lambda>\<Sigma>. embed.samequot_exec (\<epsilon> \<Sigma>)"

interpretation embed "set \<circ> (\<sigma> (\<Sigma> :: 'a :: linorder list))" \<pi> "\<epsilon> \<Sigma>"
  where "embed.lderiv (\<epsilon> \<Sigma>) = norm_lderiv \<Sigma>"
  and   "embed.rderiv (\<epsilon> \<Sigma>) = norm_rderiv \<Sigma>"
  and   "embed.rderiv_and_add (\<epsilon> \<Sigma>) = norm_rderiv_and_add \<Sigma>"
  and   "embed.samequot_exec (\<epsilon> \<Sigma>) = quot \<Sigma>"
  by (unfold_locales) (auto simp: norm_lderiv_def norm_rderiv_def norm_rderiv_and_add_def quot_def \<sigma>_def \<pi>_def \<epsilon>_def)

definition norm_step' where [code del]:
  "norm_step' \<equiv> \<lambda>\<Sigma>. equivalence_checker.step' (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (Smart_Constructors_Normalization.norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_closure' where [code del]:
  "norm_closure' \<equiv> \<lambda>\<Sigma>. equivalence_checker.closure' (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (Smart_Constructors_Normalization.norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_check_eqv' where [code del]:
  "norm_check_eqv' \<equiv> \<lambda>\<Sigma>. equivalence_checker.check_eqv' (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (Smart_Constructors_Normalization.norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_step where [code del]:
  "norm_step \<equiv> \<lambda>\<Sigma>. equivalence_checker.step (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (Smart_Constructors_Normalization.norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_closure where [code del]:
  "norm_closure \<equiv> \<lambda>\<Sigma>. equivalence_checker.closure (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (Smart_Constructors_Normalization.norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_check_eqv where [code del]:
  "norm_check_eqv \<equiv> \<lambda>\<Sigma>. equivalence_checker.check_eqv (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (Smart_Constructors_Normalization.norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"
definition norm_check_eqv_counterexample where [code del]:
  "norm_check_eqv_counterexample \<equiv> \<lambda>\<Sigma>. equivalence_checker.check_eqv_counterexample (\<sigma> \<Sigma>) (\<epsilon> \<Sigma>) (Smart_Constructors_Normalization.norm :: 'a::linorder T rexp \<Rightarrow> 'a T rexp)"

lemmas norm_defs = wf_rexp_def
  norm_check_eqv_def norm_closure_def norm_step_def norm_check_eqv_counterexample_def
  norm_check_eqv'_def norm_closure'_def norm_step'_def 

interpretation norm: equivalence_checker "\<sigma> \<Sigma>" \<pi> "\<epsilon> \<Sigma>" Smart_Constructors_Normalization.norm "\<LL> \<Sigma>"
  where "norm.check_eqv' = norm_check_eqv' \<Sigma>"
  and   "norm.check_eqv = norm_check_eqv \<Sigma>"
  and   "norm.check_eqv_counterexample = norm_check_eqv_counterexample \<Sigma>"
  and   "norm.closure' = norm_closure' \<Sigma>"
  and   "norm.closure = norm_closure \<Sigma>"
  and   "norm.step' = norm_step' \<Sigma>"
  and   "norm.step = norm_step \<Sigma>"
  by unfold_locales (auto simp: norm_defs trans[OF lang_norm[OF iffD2[OF ACI_norm_wf]] ACI_norm_lang])


abbreviation "ext \<Sigma> \<equiv> None # map Some (\<Sigma> :: 'a :: linorder list)"

definition any where [code del]:
  "any \<equiv> \<lambda>\<Sigma>. formula.any (ext \<Sigma>)"
definition pre_wf_formula where [code del]:
  "pre_wf_formula \<equiv> \<lambda>\<Sigma>. formula.pre_wf_formula (ext \<Sigma>)"
definition wf_formula where [code del]:
  "wf_formula \<equiv> \<lambda>\<Sigma>. formula.wf_formula (ext \<Sigma>)"
definition valid_ENC where [code del]: "valid_ENC \<equiv> \<lambda>\<Sigma>. formula.valid_ENC (ext \<Sigma>)"
definition ENC where [code del]: "ENC \<equiv> \<lambda>\<Sigma>. formula.ENC (ext \<Sigma>)"
definition rexp_of where [code del]: "rexp_of \<equiv> \<lambda>\<Sigma>. formula.rexp_of (ext \<Sigma>)"
definition rexp_of_alt where [code del]: "rexp_of_alt \<equiv> \<lambda>\<Sigma>. formula.rexp_of_alt (ext \<Sigma>)"
definition rexp_of' where [code del]: "rexp_of' \<equiv> \<lambda>\<Sigma>. formula.rexp_of' (ext \<Sigma>)"

lemmas formula_defs = pre_wf_formula_def wf_formula_def any_def
  rexp_of_def rexp_of'_def rexp_of_alt_def ENC_def valid_ENC_def FOV_def SOV_def

interpretation \<Phi>: formula "ext (\<Sigma> :: 'a :: linorder list)"
  where "alphabet.wf (set o \<sigma> \<Sigma>) = wf_rexp \<Sigma>"
  and   "embed.rderiv (\<epsilon> \<Sigma>) = norm_rderiv \<Sigma>"
  and   "embed.rderiv_and_add (\<epsilon> \<Sigma>) = norm_rderiv_and_add \<Sigma>"
  and   "embed.samequot_exec (\<epsilon> \<Sigma>) = quot \<Sigma>"
  and   "\<Phi>.any = any \<Sigma>"
  and   "\<Phi>.pre_wf_formula = pre_wf_formula \<Sigma>"
  and   "\<Phi>.wf_formula = wf_formula \<Sigma>"
  and   "\<Phi>.rexp_of = rexp_of \<Sigma>"
  and   "\<Phi>.rexp_of_alt = rexp_of_alt \<Sigma>"
  and   "\<Phi>.rexp_of' = rexp_of' \<Sigma>"
  and   "\<Phi>.valid_ENC = valid_ENC \<Sigma>"
  and   "\<Phi>.ENC = ENC \<Sigma>"
  by unfold_locales
    (auto simp: \<sigma>_def \<pi>_def wf_rexp_def norm_rderiv_def norm_rderiv_and_add_def quot_def formula_defs)

definition check_eqv where
"check_eqv \<Sigma> n \<phi> \<psi> \<longleftrightarrow> wf_formula \<Sigma> n (FOr \<phi> \<psi>) \<and>
   norm_check_eqv' (ext \<Sigma>) n (rexp_of \<Sigma> n (norm \<phi>)) (rexp_of \<Sigma> n (norm \<psi>))"

definition check_eqv_counterexample where
"check_eqv_counterexample \<Sigma> n \<phi> \<psi> =
   norm_check_eqv_counterexample (ext \<Sigma>) n (rexp_of \<Sigma> n (norm \<phi>)) (rexp_of \<Sigma> n (norm \<psi>))"

definition check_eqv' where
"check_eqv' \<Sigma> n \<phi> \<psi> \<longleftrightarrow> wf_formula \<Sigma> n (FOr \<phi> \<psi>) \<and>
   norm_check_eqv' (ext \<Sigma>) n (rexp_of' \<Sigma> n (norm \<phi>)) (rexp_of' \<Sigma> n (norm \<psi>))"


lemmas lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of_norm = trans[OF sym[OF \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_norm] \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of]

lemma soundness: "check_eqv \<Sigma> n \<phi> \<psi> \<Longrightarrow> \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S \<Sigma> n \<phi> = \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S \<Sigma> n \<psi>"
  by (rule box_equals[OF norm.soundness'
  sym[OF trans[OF lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of_norm]] sym[OF trans[OF lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of_norm]]])
  (auto simp: check_eqv_def split: sum.splits option.splits)

lemmas lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of'_norm = trans[OF sym[OF \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_norm] \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of']

lemma soundness': "check_eqv' \<Sigma> n \<phi> \<psi> \<Longrightarrow> \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S \<Sigma> n \<phi> = \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S \<Sigma> n \<psi>"
  by (rule box_equals[OF norm.soundness'
  sym[OF trans[OF lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of'_norm]] sym[OF trans[OF lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of'_norm]]])
   (auto simp: check_eqv'_def split: sum.splits option.splits)

lemma completeness:
assumes "\<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S \<Sigma> n \<phi> = \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S \<Sigma> n \<psi>" "wf_formula \<Sigma> n (FOr \<phi> \<psi>)"
shows "check_eqv \<Sigma> n \<phi> \<psi>"
  using assms(2) unfolding check_eqv_def
  by (intro conjI[OF assms(2) norm.completeness',
                OF box_equals[OF assms(1) lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of_norm lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of_norm]])
   (auto split: sum.splits option.splits intro!: \<Phi>.wf_rexp_of)

lemma completeness':
assumes "\<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S \<Sigma> n \<phi> = \<Phi>.lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S \<Sigma> n \<psi>" "wf_formula \<Sigma> n (FOr \<phi> \<psi>)"
shows "check_eqv' \<Sigma> n \<phi> \<psi>"
  using assms(2) unfolding check_eqv'_def
  by (intro conjI[OF assms(2) norm.completeness',
                OF box_equals[OF assms(1) lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of'_norm lang\<^isub>W\<^isub>S\<^isub>1\<^isub>S_rexp_of'_norm]])
   (auto split: sum.splits option.splits intro!: \<Phi>.wf_rexp_of')

end
